## (c) Barbara Eckl-Ganser, February 2019 - October 2020
## ecklbarb@fim.uni-passau.de
## Analysis for hierarchy structures

## SOURCE NECESSARY SCRIPTS (UTIL AND PLOT)
source("coronet/util-init.R", chdir = TRUE) # mandatory
source("configuration.R") # mandatory
source("hierarchy_util.R")
source("hierarchy_additions.R")

## LOAD NEEDED LIBRARIES

#library("igraph")
library("ggplot2")
library("plyr")
#library("gplots")
#library("gridExtra")
#library("reshape2")

## CONFIGURE PARAMETERS ONLY NEEDED FOR THIS SCRIPT
ANONYMOUS = FALSE # not needed if the input data is already anonymized

simplifyNetworks = TRUE

#AUTHOR.RELATION = "issue" #"issue" #"mail", "cochangeMail", "cochangeIssue", "cochange"

# convert author relation to R vector of relations:
AUTHOR.RELATION = switch(AUTHOR.RELATION,
                         "cochangeMail" = c("cochange", "mail"),
                         "cochangeIssue" = c("cochange", "issue"),
                         AUTHOR.RELATION)

if (length(AUTHOR.RELATION) > 1) {
  SPLIT.BASIS = "commits"
  if (length(AUTHOR.RELATION) == 2 & "cochange" %in% AUTHOR.RELATION & "mail" %in% AUTHOR.RELATION) {
    author.relation.for.file.name = "CochangeMail"
    FIRST.ACTIVITY.TYPE = c("mails", "commits")
  } else if (length(AUTHOR.RELATION) == 2 & "cochange" %in% AUTHOR.RELATION & "issue" %in% AUTHOR.RELATION) {
    author.relation.for.file.name = "CochangeIssue"
    FIRST.ACTIVITY.TYPE = c("issues", "commits")
  } else {
    author.relation.for.file.name = "multipleRelations" # to be specificed if needed
    FIRST.ACTIVITY.TYPE = c() # to be specified if needed
  }
} else if (AUTHOR.RELATION == "mail") {
  SPLIT.BASIS = "mails"
  author.relation.for.file.name = AUTHOR.RELATION
  FIRST.ACTIVITY.TYPE = c("mails")
} else if (AUTHOR.RELATION == "issue") {
  SPLIT.BASIS = "issues"
  author.relation.for.file.name = AUTHOR.RELATION
  FIRST.ACTIVITY.TYPE = c("issues")
} else {
  SPLIT.BASIS = "commits"
  author.relation.for.file.name = AUTHOR.RELATION
  FIRST.ACTIVITY.TYPE = c("commits")
}

#TIME.PERIOD = "6 months"
#SLIDING.WINDOW = TRUE
logging::loginfo(paste0("SLIDING.WINDOW: ", SLIDING.WINDOW, ", TIME.PERIOD: ", TIME.PERIOD))
OUTPUT.PATH = file.path(OUTPUT.PATH, "hierarchy-anonymized")

dir.create(OUTPUT.PATH, showWarnings = FALSE, recursive = TRUE)

logging::loginfo(paste("AUTHOR.RELATION: ", AUTHOR.RELATION, ", TIME.PERIOD: ", TIME.PERIOD, sep=" "))

## / / / / / / / / / / / / / / / / / / / / / / / / / / / / / / / / / / / / /
## CONFIGURE AND RETRIEVE NETWORK(S) ---------------------------------------

logging::loginfo("Starting configure and build issue networks.")

# initialize project configuration
proj.conf = ProjectConf$new(DATA.PATH, ANALYSIS.RANGE.TYPE, CASESTUDY, ARTIFACT)
#proj.conf$update.value("artifact.filter.base", FALSE)
proj.conf$update.value("filter.bots", TRUE)

## get ranges
ranges = proj.conf$get.value("ranges")
proj.data.whole <- ProjectData$new(project.conf = proj.conf)

unfiltered.issues = proj.data.whole$get.issues()
filtered.issues = unfiltered.issues[unfiltered.issues[["author.name"]] != "Deleted user"
                                    & unfiltered.issues[["author.name"]] != "Deleted User"
                                    & unfiltered.issues[["author.name"]] != "deleted user", ]
proj.data.whole$set.issues(filtered.issues)

logging::loginfo("Cuting data to same date finished.")

logging::loginfo("Configuration for casestuy finished.")

## all networks ---------------------------------------------------------
## initialize network configuration

net.conf.issue = NetworkConf$new()
net.conf.issue$update.values(updated.values = list(author.relation = AUTHOR.RELATION,
                                                   artifact.relation = ARTIFACT.RELATION,
                                                   simplify=simplifyNetworks))
network.builder.issue <- NetworkBuilder$new(proj.data.whole, net.conf.issue)
output.path.project = paste(OUTPUT.PATH, CASESTUDY, gsub(" ", "", TIME.PERIOD), author.relation.for.file.name, sep="/")
output.path.networks = paste(output.path.project, "networks/", sep="/")
dir.create(output.path.networks, showWarnings = FALSE, recursive = TRUE)


dump.path <- paste(output.path.networks,
                   paste(CASESTUDY, author.relation.for.file.name, sub(" ", "", TIME.PERIOD), "sliWindow",
                         SLIDING.WINDOW,"whole-network.rdata", sep = "-"),
                   sep="")
save.and.load("whole.network", dump.path, function(){
  whole.network <- network.builder.issue$get.author.network()
  return(whole.network)
}, skip = FALSE)

node.degree <- igraph::degree(whole.network, mode="all")
snd <- sort(node.degree, decreasing = TRUE)
SELECTED.DEVELOPERS <- names(head(snd, 10))
# SELECTED.DEVELOPERS <- SELECTED.DEVELOPERS[!grepl("=",SELECTED.DEVELOPERS)]

set.seed(2389)
snd10 = snd[snd > 10]     # greater node degree for random selected nodes than ten
random.numbers <- floor(runif(10, min=0, max=length(snd10)))
RANDOM.SELECTED.DEVELOPERS <- names(snd10)[random.numbers]
# NOTICE: This will be updated below if one of these developers does not appear in hierarchy data (e.g., degree <= 1)


split.data.and.build.networks <- function() {
  cf.data.all.issue <- split.data.time.based(proj.data.whole, time.period = TIME.PERIOD,
                                             split.basis = SPLIT.BASIS, sliding.window = SLIDING.WINDOW)

  ## author network
  networks <- parallel::mclapply(cf.data.all.issue, function(d) {
    network.builder <- NetworkBuilder$new(d, net.conf.issue)
    network.builder$update.network.conf(updated.values = list(author.relation = AUTHOR.RELATION))
    return(network.builder$get.author.network())
  })
  return(networks)
}


dump.path <- paste(output.path.networks,
                   paste(CASESTUDY, author.relation.for.file.name, sub(" ", "", TIME.PERIOD), "sliWindow",
                   SLIDING.WINDOW, "networks.rdata", sep = "-"),
                   sep="")
save.and.load("networks", dump.path, split.data.and.build.networks, skip = FALSE)

networks <- networks[lapply(networks, vcount) >1]
logging::loginfo("Finished configure and build all mail networks.")




### COMPUTE NODE DEGREE -------------------------------------------------------
node.degrees <- lapply(names(networks), function(nn){
  net <- networks[[nn]]
  num.developers <- vcount(net)
  node.degs <- igraph::degree(net, mode = "all")
  frequency.degs <- plyr::count(node.degs)

  if (0 %in% frequency.degs$x) {
    num.deg.zero <- frequency.degs[frequency.degs$x == 0, "freq"]
    num.deg.zero.percent <- 100*num.deg.zero/num.developers
  } else {
    num.deg.zero <- 0
    num.deg.zero.percent <- 0
  }

  if (1 %in% frequency.degs$x) {
    num.deg.one <- frequency.degs[frequency.degs$x == 1, "freq"]
    num.deg.one.percent <- 100*num.deg.one/num.developers
  } else {
    num.deg.one <- 0
    num.deg.one.percent <- 0
  }

  if (length(setdiff(frequency.degs$x, c(0,1))) != 0) {
    num.deg.greater.one <- sum(frequency.degs[frequency.degs$x > 1, "freq"])
    num.deg.greater.one.percent <- 100*num.deg.greater.one/num.developers
  } else {
    num.deg.greater.one <- 0
    num.deg.greater.one.percent <- 0
  }


  return(data.frame("range"=nn, "num.dev"=num.developers,
                    "num.deg.zero"=num.deg.zero, "num.deg.one"=num.deg.one, "num.deg.greater.one"=num.deg.greater.one,
                    "num.deg.zero.percent"=num.deg.zero.percent, "num.deg.one.percent"=num.deg.one.percent,
                    "num.deg.greater.one.percent"=num.deg.greater.one.percent,
                    "entire.num.dev"=length(snd)))
})

node.degrees <- rbind.fill(node.degrees)
node.degrees$range.number <- c(1: nrow(node.degrees))
rownames(node.degrees) <- node.degrees$range


## plots for node degree
# new_nd<-subset(node.degrees, select=c("range", "num.deg.zero", "num.deg.one", "num.deg.greater.one"))
# xdf <- reshape2::melt(new_nd)
#
# df_xdf <- ddply(xdf, "range",
#                    transform, label_ypos=cumsum(value))
#
# test.plot <- ggplot(data=df_xdf, aes(x=range, y=value, fill=variable)) +
#   geom_bar(stat="identity") +
#   geom_text(aes(y=label_ypos, label=value), vjust=1.6,
#             color="white") #+
  #scale_fill_brewer(palette="Paired")+
  #theme_minimal()

logging::loginfo("Computation of node degrees finished.")

## COMPUTE HIERARCHY ----------------------------------------------------------
## hierarchy
hiera <- lapply(names(networks), function(range){
  if (ecount(networks[[range]]) > 0 && vcount(networks[[range]]) > 0) {
    hier <- metrics.hierarchy(networks[[range]])
    if(nrow(hier) != 0) {
      logging::loginfo(paste("Compute hierarchy of range", range, "with", nrow(hier), "rows ", sep = " "))
      finite.hier <- subset(hier, !(is.nan(cc) | is.infinite(log.cc))) # because of new metrics.hierarchy

      logging::loginfo(paste("Compute hierarchy of range", range, "with", nrow(finite.hier), "rows ", sep = " "))
      if (nrow(finite.hier) != 0) {
        hier.df <- data.frame("range" = range, "hierarchy" = finite.hier)
        return(hier.df)
      } else {
        return(NULL)
      }
    } else {
      return(NULL)
    }
  } else {
    return(NULL)
  }
})

names(hiera) <- names(networks)
for(i in 1:length(names(hiera))) {
  hierarchy <- hiera[[i]]
  hierarchy$range.number <- i
  hierarchy$author.name <- rownames(hierarchy)
  hiera[[i]] <- hierarchy
}

hiera <- plyr::compact(hiera)

hiera.bound.rows = dplyr::bind_rows(hiera)
write.csv(hiera.bound.rows, file = paste0(OUTPUT.PATH, "/" , CASESTUDY, "/" ,author.relation.for.file.name ,"hierarchy-data.csv"))

logging::loginfo("Computation of hierarchy finished.")


## Update randomly selected developers if one does not appear in hiearchy data (e.g., only having degree <= 1)

random.devs.to.remove = list()
new.random.devs = list()
for (random.dev in RANDOM.SELECTED.DEVELOPERS) {
  if (! (random.dev %in% hiera.bound.rows[["author.name"]])) {
    random.dev.new = random.dev
    index = which(names(snd10) == random.dev.new)
    while (random.dev.new %in% RANDOM.SELECTED.DEVELOPERS || random.dev.new %in% new.random.devs
           || !(random.dev.new %in% hiera.bound.rows[["author.name"]])) {
      index = index + 1
      random.dev.new <<- names(snd10)[index]
      if (is.null(random.dev.new) || length(random.dev.new) == 0 || is.na(random.dev.new)) {
          index = 1
        random.dev.new <<- names(snd10)[index]
      }
    }
    random.devs.to.remove = c(random.devs.to.remove, random.dev)
    new.random.devs = c(new.random.devs, random.dev.new)
  }
}
for (i in seq_along(random.devs.to.remove)) {
  RANDOM.SELECTED.DEVELOPERS = replace(RANDOM.SELECTED.DEVELOPERS, RANDOM.SELECTED.DEVELOPERS == random.devs.to.remove[[i]], new.random.devs[[i]])
}


## Hierarchy plots with horizontal lines -----------------------------------------

# one output file per plot
output.path.hierarchical.structure = paste(output.path.project, "annotation/single-plots/" ,sep="/")
dir.create(output.path.hierarchical.structure, showWarnings = FALSE, recursive = TRUE)

plots.annotation <- list()
i = 1


for (range.name in names(hiera)) {
  if (!is.null(ncol(hiera[[range.name]])) && nrow(hiera[[range.name]]) > 0) {
    logging::loginfo(paste0("Number of rows in the hierarchy:", nrow(hiera[[range.name]])) )
    plots.annotation[[i]] <- plot.hierarchy.only.realistic.values(hiera[[range.name]],
                                                                  number.of.steps = NUMBER.OF.STEPS,
                                                                  all.lines = FALSE,
                                                                  all.lines.colored = TRUE,
                                                                  with.left.group = FALSE,
                                                                  only.selected.group = FALSE,
                                                                  vertical.lines = TRUE)
    plots.annotation[[i]] <- plots.annotation[[i]] + ggtitle(paste0(hiera[[range.name]]$range.number[1],
                                                                    ": ", paste(substr(range.name,1,10),
                                                                                substr(range.name,21,30), sep=" to ")))
    ## save plot
    plot.output.file <- paste(output.path.hierarchical.structure, CASESTUDY,
                              paste0(hiera[[range.name]]$range.number[1], ":",
                                     paste(substr(range.name,1,10), substr(range.name,21,30), sep="to")),
                              ".png", sep="-")

    ggsave(filename = plot.output.file, plot = plots.annotation[[i]], limitsize = FALSE)
  } else {
    i = i-1
  }


  i = i+1
}


output.path.annotation = paste(output.path.project, "annotation/" ,sep="/")
dir.create(output.path.annotation, showWarnings = FALSE, recursive = TRUE)

selected.plots <- sample(plots.annotation, 10)

plot.output.file <- paste(output.path.annotation,
                          paste(CASESTUDY, author.relation.for.file.name,
                                "randomly-selected-hierarchy-plots-for-annotation.pdf", sep = "-"),
                          sep="")

pdf(file=plot.output.file, height= 6*ceiling(length(selected.plots)/3), width = 10*3)
do.call("grid.arrange", c(selected.plots, ncol=3, newpage=TRUE))
dev.off()


plot.output.file <- paste(output.path.annotation,
                          paste(CASESTUDY, author.relation.for.file.name,
                                "hierarchy-plots-with-files-for-annotation.pdf", sep = "-"),
                          sep="")

pdf(file=plot.output.file, height= 6*ceiling(length(plots.annotation)/3), width = 10*3)
do.call("grid.arrange", c(plots.annotation, ncol=3, newpage=TRUE))
dev.off()


# colored.lines <- lapply(hiera, function(h.df) {
#   return(plot.hierarchy.only.realistic.values(h.df,
#                                               number.of.steps = NUMBER.OF.STEPS,
#                                               all.lines = FALSE,
#                                               all.lines.colored = TRUE,
#                                               with.left.group = FALSE,
#                                               only.selected.group = FALSE,
#                                               vertical.lines = TRUE))
# })
#

logging::loginfo("Plotting of hierarchy with vertical lines for manual annotation finished.")

## DETECTION OF HIERARCHICAL STRUCTURE ----------------------------------------

step.wise.line <- lapply(hiera, compute.hierarchical.parts) # possibly returns NULL values!
step.wise.line <- plyr::compact(step.wise.line)
marked.data <- mark.realistic.values.of.whole.project.history(step.wise.line)
data.with.normalized.criteria <- normalize.criteria(marked.data)

### with previous selection --------------------
selected.group.with.pre.selection <- maximize.criteria.sum(data.with.normalized.criteria,  previous.selection = TRUE)


# numbers.of.selected.candidates <- extract.number.of.selected.candidate.group(selected.group.with.pre.selection)
plots3 <- list()
i = 1
for (range.name in names(selected.group.with.pre.selection)) {
  plots3[[i]] <- plot.hierarchy.only.for.selected.group(hiera[range.name],
                                                        selected.group.with.pre.selection[range.name],
                                                        node.degrees[range.name,])
  plots3[[i]] <- plots3[[i]] + ggtitle(paste0(hiera[[range.name]]$range.number[1],
                                              ": ", paste(substr(range.name,1,10),
                                                          substr(range.name,21,30), sep=" to ")))
  #+ geom_text(aes(y=0), x=1, label=i, vjust = 1.2, size=3)

  i = i+1
}

output.path.hierarchical.structure = paste(output.path.project, "hierarchical-structure/" ,sep="/")
dir.create(output.path.hierarchical.structure, showWarnings = FALSE, recursive = TRUE)

plot.output.file <- paste(output.path.hierarchical.structure,
                          paste(CASESTUDY, author.relation.for.file.name,
                                "selected_hierarchical-structure-weight3-changedHierarchy.pdf", sep = "-"),
                          sep="")

pdf(file=plot.output.file, height= 6*ceiling(length(plots3)/5), width = 10*5)
do.call("grid.arrange", c(plots3, ncol=5, newpage=TRUE))
dev.off()

#
output.path.hierarchical.structure = paste(output.path.project, "hierarchical-structure/single-plots/" ,sep="/")
dir.create(output.path.hierarchical.structure, showWarnings = FALSE, recursive = TRUE)

logging::loginfo("Selection of a hierarchical part and its plotting finished.")

## ANALYSING THRESHOLDS -------------------------------------------------------

marked.data.with.threshold <- rbind.fill(marked.data)
part.of.variance = 1 - length(which(marked.data.with.threshold$variance.smaller.threshold)) /nrow(marked.data.with.threshold)
part.of.percentage = 1 - length(which(marked.data.with.threshold$percent.n.higher.threshold)) /nrow(marked.data.with.threshold)
part.of.angle = 1 - length(which(marked.data.with.threshold$angel.smaller.threshold)) /nrow(marked.data.with.threshold)

part.of.possible = length(which(marked.data.with.threshold$possible.regression.line)) /nrow(marked.data.with.threshold)

## analyze criteria --------------

marked.data.with.weighted.sum <- mapply(function(criteria.data, candidates){
  maximum.selected.weighted.sum <- candidates[candidates$selected.value, "weighted.sum"]
  criteria <- criteria.data[criteria.data$weighted.sum > maximum.selected.weighted.sum[[1]], ]
  if (!is.null(criteria) && nrow(criteria) > 0) {
    variance <- any(!criteria$variance.smaller.threshold) # ich verstehe das Ergebnis nicht
    angle <- any(!criteria$angel.smaller.threshold)
    percent.n <- any(!criteria$percent.n.higher.threshold)
    return(data.frame("var"=variance, "ang"=angle, "per.n"=percent.n))
  }

}, data.with.normalized.criteria, selected.group.with.pre.selection, SIMPLIFY = FALSE)

marked.data.with.analyzed.criteria <- rbind.fill(compact(marked.data.with.weighted.sum))

sum(marked.data.with.analyzed.criteria$var) # hier werden True-Werte gezählt
sum(marked.data.with.analyzed.criteria$ang)
sum(marked.data.with.analyzed.criteria$per.n)


### without previous selection ------------------
selected.group <- maximize.criteria.sum(data.with.normalized.criteria,  previous.selection = FALSE)

plots2 <- list()
i = 1
for (range.name in names(selected.group)) {
  plots2[[i]] <- plot.hierarchy.only.for.selected.group(hiera[range.name],
                                                        selected.group[range.name],
                                                        node.degrees[range.name,])
  plots2[[i]] <- plots2[[i]] + ggtitle(paste0(hiera[[range.name]]$range.number[1], ": ", paste(substr(range.name,1,10), substr(range.name,21,30), sep=" to ")))
  #+ geom_text(aes(y=0), x=1, label=i, vjust = 1.2, size=3)
  i = i+1
}

output.path.hierarchical.structure = paste(output.path.project, "hierarchical-structure/" ,sep="/")
dir.create(output.path.hierarchical.structure, showWarnings = FALSE, recursive = TRUE)

plot.output.file <- paste(output.path.hierarchical.structure,
                          paste(CASESTUDY, author.relation.for.file.name,
                                "selected_hierarchical-structure-weight3-without-previous-selection.pdf", sep = "-"),
                          sep="")

pdf(file=plot.output.file, height= 6*ceiling(length(plots2)/5), width = 10*5)
do.call("grid.arrange", c(plots2, ncol=5, newpage=TRUE))
dev.off()

logging::loginfo("Further analysis regarding thresholds finished.")


## HIERARCHY CRITERIA ---------------------------------------------------------

extracted.groups <- select.candidate.partial.hierarchy(selected.group.with.pre.selection)
extracted.df <- rbind.fill(extracted.groups)

if (!is.null(extracted.df)) {
  for (i in seq_along(names(extracted.groups))) {
    extracted.df[i, "range.number"] <- hiera[[names(extracted.groups)[i]]]$range.number[1]
  }

  output.path.aggregation = paste(output.path.project, "aggregation-hierarchy-criteria/" ,sep="/")
  dir.create(output.path.aggregation, showWarnings = FALSE, recursive = TRUE)
  plot.output.file <- paste(output.path.aggregation,
                            paste(CASESTUDY, author.relation.for.file.name,
                                  "aggregation_criteria-weight3changedHierarchy.pdf", sep = "-"),
                            sep="")
  plot.parameters.from.one.project(extracted.df, node.degrees, plot.output.file)

  logging::loginfo("Plotting of hierarchy finished.")
}


## ANALYSE OF SELECTED DEVELOPERS ---------------------------------------------

## Movement of one developer in hierarchy ---------
max.bin.size <- compute.maximal.bin.size.for.movement(hiera, SELECTED.DEVELOPERS)

output.path.movement = paste(output.path.project, "movement/", sep="/")
dir.create(output.path.movement, showWarnings = FALSE, recursive = TRUE)

name.iterator = 1
for (developer.name in SELECTED.DEVELOPERS) {
  logging::loginfo(paste("Starting plotting movement of ",developer.name, ", ", name.iterator))

  if (ANONYMOUS) {
    plot.output.file <- paste(output.path.movement,
                              paste(CASESTUDY, author.relation.for.file.name, paste("developer", name.iterator, sep=""),
                                    "hierarchy_movement.pdf", sep = "-"),
                              sep="")
  } else {
    plot.output.file <- paste(output.path.movement,
                            paste(CASESTUDY, author.relation.for.file.name, gsub(" ", "", developer.name),
                                  "hierarchy_movement.pdf", sep = "-"),
                            sep="")
  }

  plot.movement.in.hierarchy(hiera, c(developer.name), plot.output.file, max.bin.size)

  if (ANONYMOUS) {
    plot.output.file <- paste(output.path.movement,
                              paste(CASESTUDY, author.relation.for.file.name, paste("developer", name.iterator, sep=""),
                                    "hierarchy_movement_combined.pdf", sep = "-"),
                              sep="")
  } else {
    plot.output.file <- paste(output.path.movement,
                              paste(CASESTUDY, author.relation.for.file.name, gsub(" ", "", developer.name),
                                    "hierarchy_movement_combined.pdf", sep = "-"),
                              sep="")
  }

  plot.movement.in.hierarchy.combination(hiera, c(developer.name), plot.output.file, max.bin.size)
  logging::loginfo(paste("Finished plotting movement of ",developer.name, ", ", name.iterator))

  name.iterator <- name.iterator + 1
}
logging::loginfo("Finished plotting movement.")

## Movement: random selected --------------------------------------------------

output.path.movement = paste(output.path.project, "movement/", sep="/")
dir.create(output.path.movement, showWarnings = FALSE, recursive = TRUE)

name.iterator = 100
for (developer.name in RANDOM.SELECTED.DEVELOPERS) {
  logging::loginfo(paste("Starting plotting movement of ",developer.name, ", ", name.iterator))

  if (ANONYMOUS) {
    plot.output.file <- paste(output.path.movement,
                              paste(CASESTUDY, author.relation.for.file.name, paste("developer", name.iterator, sep=""),
                                    "hierarchy_movement_random.pdf", sep = "-"),
                              sep="")
  } else {
    plot.output.file <- paste(output.path.movement,
                              paste(CASESTUDY, author.relation.for.file.name, gsub(" ", "", developer.name),
                                    "hierarchy_movement_random.pdf", sep = "-"),
                              sep="")
  }

  plot.movement.in.hierarchy(hiera, c(developer.name), plot.output.file, max.bin.size)

  if (ANONYMOUS) {
    plot.output.file <- paste(output.path.movement,
                              paste(CASESTUDY, author.relation.for.file.name, paste("developer", name.iterator, sep=""),
                                    "hierarchy_movement_combined_random.pdf", sep = "-"),
                              sep="")
  } else {
    plot.output.file <- paste(output.path.movement,
                              paste(CASESTUDY, author.relation.for.file.name, gsub(" ", "", developer.name),
                                    "hierarchy_movement_combined_random.pdf", sep = "-"),
                              sep="")
  }

  plot.movement.in.hierarchy.combination(hiera, c(developer.name), plot.output.file, max.bin.size)
  logging::loginfo(paste("Finished plotting movement of ",developer.name, ", ", name.iterator))

  name.iterator <- name.iterator + 1
}
logging::loginfo("Finished plotting movement.")

## Neighborhood in hierarchy --------------------------------------------------
output.path.neighborhood = paste(output.path.project, "neighborhood/" ,sep="/")
dir.create(output.path.neighborhood, showWarnings = FALSE, recursive = TRUE)

name.iterator <- 1
for (developer.name in SELECTED.DEVELOPERS) {
  logging::loginfo(paste("Starting plotting neighborhood of ", developer.name))
  if (ANONYMOUS) {
    plot.output.file <- paste(output.path.neighborhood,
                              paste(CASESTUDY, author.relation.for.file.name, paste("developer", name.iterator, sep=""),
                                    "neighborhood.pdf", sep = "-"),
                              sep="")
  } else {
    plot.output.file <- paste(output.path.neighborhood,
                              paste(CASESTUDY, author.relation.for.file.name, gsub(" ", "", developer.name),
                                    "neighborhood.pdf", sep = "-"),
                              sep="")
  }
  plot.neighborhood.of.selected.developer(developer.name, hiera, networks, plot.output.file)
  logging::loginfo(paste("Finished plotting neighborhood of ",developer.name))
  name.iterator <- name.iterator + 1
}
logging::loginfo("Finished plotting neighborhood.")

## Aggregated Neighborhood -----------------------------------------------------

output.path.ag.neighborhood = paste(output.path.project, "aggregated-neighborhood/" ,sep="/")
dir.create(output.path.ag.neighborhood, showWarnings = FALSE, recursive = TRUE)

name.iterator <- 1
for (developer.name in SELECTED.DEVELOPERS) {
  logging::loginfo(paste("Starting plotting aggregated neighborhood of ", developer.name))
  if (ANONYMOUS) {
    plot.output.file <- paste(output.path.ag.neighborhood,
                              paste(CASESTUDY, author.relation.for.file.name, paste("developer", name.iterator, sep=""),
                                    "ag_neighborhood.pdf", sep = "-"),
                              sep="")
  } else {
    plot.output.file <- paste(output.path.ag.neighborhood,
                              paste(CASESTUDY, author.relation.for.file.name, gsub(" ", "", developer.name),
                                    "ag_neighborhood.pdf", sep = "-"),
                              sep="")
  }
  plot.aggregated.neighborhood.of.selected.developer(developer.name, hiera, networks, selected.group,
                                                     number.of.steps = NUMBER.OF.STEPS,
                                                     plot.path = plot.output.file)
  logging::loginfo(paste("Finished plotting aggregated neighborhood of ",developer.name))
  name.iterator <- name.iterator + 1
}
logging::loginfo("Finished plotting neighborhood.")

## Aggregated Neighborhood of random developers
name.iterator <- 1
for (developer.name in RANDOM.SELECTED.DEVELOPERS) {
  logging::loginfo(paste("Starting plotting aggregated neighborhood of ", developer.name))
  if (ANONYMOUS) {
    plot.output.file <- paste(output.path.ag.neighborhood,
                              paste(CASESTUDY, author.relation.for.file.name, paste("developer", name.iterator, sep=""),
                                    "ag_neighborhood_random.pdf", sep = "-"),
                              sep="")
  } else {
    plot.output.file <- paste(output.path.ag.neighborhood,
                              paste(CASESTUDY, author.relation.for.file.name, gsub(" ", "", developer.name),
                                    "ag_neighborhood_random.pdf", sep = "-"),
                              sep="")
  }
  plot.aggregated.neighborhood.of.selected.developer(developer.name, hiera, networks, selected.group,
                                                     number.of.steps = NUMBER.OF.STEPS,
                                                     plot.path = plot.output.file)
  logging::loginfo(paste("Finished plotting aggregated neighborhood of ",developer.name))
  name.iterator <- name.iterator + 1
}
logging::loginfo("Finished plotting neighborhood.")

# Random Neighborhood in hierarchy --------------------------------------------------
output.path.neighborhood = paste(output.path.project, "neighborhood/" ,sep="/")
dir.create(output.path.neighborhood, showWarnings = FALSE, recursive = TRUE)

name.iterator <- 100
for (developer.name in RANDOM.SELECTED.DEVELOPERS) {
  logging::loginfo(paste("Starting plotting neighborhood of ", developer.name))
  if (ANONYMOUS) {
    plot.output.file <- paste(output.path.neighborhood,
                              paste(CASESTUDY, author.relation.for.file.name, paste("developer", name.iterator, sep=""),
                                    "neighborhood_random.pdf", sep = "-"),
                              sep="")
  } else {
    plot.output.file <- paste(output.path.neighborhood,
                              paste(CASESTUDY, author.relation.for.file.name, gsub(" ", "", developer.name),
                                    "neighborhood_random.pdf", sep = "-"),
                              sep="")
  }
  plot.neighborhood.of.selected.developer(developer.name, hiera, networks, plot.output.file)
  logging::loginfo(paste("Finished plotting neighborhood of ",developer.name))
  name.iterator <- name.iterator + 1
}
logging::loginfo("Finished plotting neighborhood.")

## CORE/PERIPHERAL CLASSIFICATION ---------------------------------------------
## based on eigenvector centrality

logging::loginfo("Starts plotting hierarchy with core/peripheral classification.")
output.path.core.peripheral = paste(output.path.project, "core-peripheral-classification/" ,sep="/")
dir.create(output.path.core.peripheral, showWarnings = FALSE, recursive = TRUE)
plot.output.file <- paste(output.path.core.peripheral,
                          paste(CASESTUDY, author.relation.for.file.name,"hierarchy_marked_core_developers.pdf", sep = "-"),
                          sep="")
plot.with.marked.core.developers(hiera, networks, node.degrees, plot.output.file)
logging::loginfo("Finished plotting hierarchy with core/peripheral classification.")

## stay duration --------------------------------------------------------------
logging::loginfo("Starting time since first activity.")
output.path.first.activity = paste(output.path.project, "time-since-first-activity/" ,sep="/")
dir.create(output.path.first.activity, showWarnings = FALSE, recursive = TRUE)
plot.output.file <- paste(output.path.first.activity,
                          paste(CASESTUDY, author.relation.for.file.name,"time-since-first-activity.pdf", sep = "-"),
                          sep="")
hierarchy.plot.with.stay.period(networks, proj.data.whole, FIRST.ACTIVITY.TYPE,
                                hiera, plot.output.file)
logging::loginfo("Finished time since first activity.")

## Aggregation of tenure ------------------------------------------------

logging::loginfo("Starting aggregation of tenure")
hierarchy.with.tenure.and.hierarchy <- compute.hierarchy.and.tenure(networks, proj.data.whole,
                                                                    FIRST.ACTIVITY.TYPE, hiera,
                                                                    selected.group.with.pre.selection)

mean.tenure <- compute.mean.tenure(hierarchy.with.tenure.and.hierarchy)

median.tenure <- compute.median.tenure(hierarchy.with.tenure.and.hierarchy)

output.path.first.activity = paste(output.path.project, "aggregated-tenure/" ,sep="/")
dir.create(output.path.first.activity, showWarnings = FALSE, recursive = TRUE)

if (!all(is.na(mean.tenure$mean.tenure))) {
  plot.output.file <- paste(output.path.first.activity,
                            paste(CASESTUDY, author.relation.for.file.name,"mean-tenure-by-group.pdf", sep = "-"),
                            sep="")
  plot.hierarchy.and.tenure(mean.tenure, plot.output.file)

  plot.output.file <- paste(output.path.first.activity,
                            paste(CASESTUDY, author.relation.for.file.name,"mean-tenure-by-group-violin.pdf", sep = "-"),
                            sep="")
  plot.hierarchy.and.tenure.violin(mean.tenure, plot.output.file)

  plot.output.file <- paste(output.path.first.activity,
                            paste(CASESTUDY, author.relation.for.file.name,"tenure-by-group-over-all-ranges.pdf", sep = "-"),
                            sep="")
  plot.hierarchy.and.tenure.violin.aggregate.all.ranges(hierarchy.with.tenure.and.hierarchy, plot.output.file)

  plot.output.file <- paste(output.path.first.activity,
                            paste(CASESTUDY, author.relation.for.file.name,"tenure-per-range-all-ranges.pdf", sep = "-"),
                            sep="")
  plot.hierarchy.and.tenure.many.small.violin.in.one(hierarchy.with.tenure.and.hierarchy, plot.output.file)

  plot.output.file <- paste(output.path.first.activity,
                            paste(CASESTUDY, author.relation.for.file.name,"tenure-per-range-all-ranges-bullets.pdf", sep = "-"),
                            sep="")
  plot.hierarchy.and.tenure.many.small.violin.in.one.with.bullet(hierarchy.with.tenure.and.hierarchy,
                                                                 median.tenure ,plot.output.file)

  plot.output.file <- paste(output.path.first.activity,
                            paste(CASESTUDY, author.relation.for.file.name,"tenure-by-group-violin-per-range.pdf", sep = "-"),
                            sep="")
  plot.hierarchy.and.tenure.violin.per.range(hierarchy.with.tenure.and.hierarchy, plot.output.file)

  # statistical tests
  plot.output.file <- paste(output.path.first.activity,
                            paste(CASESTUDY, author.relation.for.file.name,"tenure-by-group-statistical-tests.pdf", sep = "-"),
                            sep="")
  compute.statistical.tests.tenure(hierarchy.with.tenure.and.hierarchy, plot.output.file)

  logging::loginfo("Finished aggregation of tenure")
} else {
  logging::loginfo("Aggregation of tenure not available.")
}

#  number of edited files ----------------------------------------------------
logging::loginfo("Starting number of edited files.")
output.path.edited.files = paste(output.path.project, "edited-files/" ,sep="/")
dir.create(output.path.edited.files, showWarnings = FALSE, recursive = TRUE)
plot.output.file <- paste(output.path.edited.files,
                          paste(CASESTUDY, author.relation.for.file.name,"number-of-edited-files.pdf", sep = "-"),
                          sep="")

hierarchy.with.arti.count <- compute.artifact.count(networks, proj.data.whole, AUTHOR.RELATION,
                                                    hiera)

plot.hierarchy.with.artifact.count(hierarchy.with.arti.count, plot.output.file)

logging::loginfo("Finished number of edited files.")

## Aggregated programming activity -----------------------------------------

logging::loginfo("Starting aggregation of programming activity")


hierarchy.artifact.count <- lapply(hierarchy.with.arti.count, rbind.fill)

hierarchy.with.edited.files.and.hierarchy <- compute.hierarchy.and.artifact.count(hierarchy.artifact.count,
                                                                                  selected.group.with.pre.selection)

mean.edited.files <- compute.mean.edited.files(hierarchy.with.edited.files.and.hierarchy)

if (!all(is.na(mean.edited.files$mean.artifact.count))) {
  output.path.first.activity = paste(output.path.project, "aggregated-edited-files/" ,sep="/")
  dir.create(output.path.first.activity, showWarnings = FALSE, recursive = TRUE)
  plot.output.file <- paste(output.path.first.activity,
                            paste(CASESTUDY, author.relation.for.file.name,"mean-edited-files-by-group.pdf", sep = "-"),
                            sep="")
  plot.hierarchy.and.edited.files(mean.edited.files, plot.output.file)

  plot.output.file <- paste(output.path.first.activity,
                            paste(CASESTUDY, author.relation.for.file.name,
                                  "edited-files-by-group-over-all-ranges.pdf", sep = "-"),
                            sep="")
  plot.hierarchy.and.edited.files.violin.aggregate.all.ranges(hierarchy.with.edited.files.and.hierarchy,
                                                              plot.output.file)
  # statistical tests
  plot.output.file <- paste(output.path.first.activity,
                            paste(CASESTUDY, author.relation.for.file.name,"edited-files-statistical-tests.pdf", sep = "-"),
                            sep="")
  compute.statistical.tests.edited.files(hierarchy.with.edited.files.and.hierarchy, plot.output.file)

  logging::loginfo("Finished aggregation of programming activity!")
} else {
  logging::loginfo("Aggregation of programming activity not available.")
}

## descriptive statistics of selected developers ------------------------------

descriptive.statistics = create.empty.data.frame(columns = c("developer", "commits", "communication.events", "active.periods",
                                                             "periods.in.hierarchy", "first.activity"))
degree.statistics = data.frame()
file.edit.statistics = data.frame()

get.descriptive.statistics.for.selected.developers <- function(selected.developers, output.file) {

  for (developer.name in selected.developers) {

    # commit count
    commits = proj.data.whole$get.commits()
    commits = commits[commits[["author.name"]] == developer.name, ]
    number.commits = length(unique(commits[["hash"]]))

    # issue/mail count
    first.activity.type.communication = FIRST.ACTIVITY.TYPE
    if (length(first.activity.type.communication) > 1) {
      if ("mails" %in% first.activity.type.communication) {
        first.activity.type.communication = "mails"
      } else if ("issues" %in% first.activity.type.communication) {
        first.activity.type.communication = "issues"
      } else {
        first.activity.type.communication = NA
      }
    }
    if (!is.na(first.activity.type.communication)) {
      getter.function = DATASOURCE.TO.ARTIFACT.FUNCTION[[first.activity.type.communication]]
      event.data = proj.data.whole[[getter.function]]()
      event.data = event.data[event.data[["author.name"]] == developer.name, ]
      id.column.name = ifelse(first.activity.type.communication == "issues", "event.id", "message.id")
      event.count = length(unique(event.data[[id.column.name]]))
    } else {
      event.count = NA
    }

    # degree
    degrees = c()
    for (network in networks) {
      network.degrees = igraph::degree(network)
      d = network.degrees[names(network.degrees) == developer.name]
      if (length(d) == 0) {
        d = NA
      }
      degrees = c(degrees, d)
    }

    # active periods
    active.periods = sum(!is.na(degrees))

    # how long active
    networks.with.first.activity = add.vertex.attribute.first.activity(networks, proj.data.whole, aggregation.level = "cumulative")
    first.activity = c()
    for (network in networks.with.first.activity) {
      devs = igraph::V(network)
      if (developer.name %in% devs$name) {
        first.activity = min(unlist(devs[devs$name == developer.name]$first.activity[[1]][FIRST.ACTIVITY.TYPE]), na.rm = TRUE)
        break
      }
    }

    # how many periods in hierarchy
    periods.in.hierarchy = 0
    for (hierarchy in hierarchy.with.tenure.and.hierarchy) {
      if ( !is.null(hierarchy) ) {
        is.part.of.hierarchy = hierarchy[hierarchy$author.name == developer.name,]$part.of.hierarchy
        if (length(is.part.of.hierarchy) > 0 && !is.na(is.part.of.hierarchy)) {
          periods.in.hierarchy = periods.in.hierarchy + is.part.of.hierarchy
        }
      }
    }

    # edited files
    edited.files = 0
    for (range in hierarchy.with.edited.files.and.hierarchy) {
      count = range[range$author.name == developer.name,]$artifact.count
      if (length(count) == 0) {
        count = 0
      }
      edited.files = c(edited.files, count)
    }

    # combine everything in one data.frame
    statistics = c(developer.name, number.commits, event.count, active.periods, periods.in.hierarchy, first.activity)
    descriptive.statistics = rbind(descriptive.statistics, statistics)
    colnames(descriptive.statistics) = c("developer.name", "number.commits", "event.count", "active.periods",
                                         "periods.in.hierarchy", "first.activity")
    degree.statistics = rbind(degree.statistics, c(developer.name, degrees))
    colnames(degree.statistics)[1] = "developer.name"
    file.edit.statistics = rbind(file.edit.statistics, c(developer.name, edited.files))
    colnames(file.edit.statistics)[1] = "developer.name"
  }

  # write to output.file
  write.csv(descriptive.statistics, paste0(output.file, ".csv"))
  save(degree.statistics, file = paste0(output.file, "_degrees.Rdata"))
  save(file.edit.statistics, file = paste0(output.file, "_editedFiles.Rdata"))
}

output.path.descriptive.statistics = paste(output.path.project, "descriptive-statistics/" ,sep="/")
dir.create(output.path.descriptive.statistics, showWarnings = FALSE, recursive = TRUE)

get.descriptive.statistics.for.selected.developers(RANDOM.SELECTED.DEVELOPERS,
                                                   file.path(output.path.descriptive.statistics,
                                                             "descriptive.statistics.random.devs"))
get.descriptive.statistics.for.selected.developers(SELECTED.DEVELOPERS,
                                                   file.path(output.path.descriptive.statistics,
                                                             "descriptive.statistics.most.active.devs"))

