## (c) Barbara Eckl-Ganser, February 2019 - October 2020
## ecklbarb@fim.uni-passau.de
## Methods to analyse hierarchy with additional information

## SOURCE NECESSARY SCRIPTS (UTIL AND PLOT)
source("coronet/util-init.R", chdir = TRUE)
source("hierarchy_util.R")

## LOAD NEEDED LIBRARIES

library("igraph")
library("ggplot2")
#library("gplots")
#library("gridExtra")
library("rcompanion")
library("effsize")

## additional parameters ---------------------------
scaleFUN <- function(x) sprintf("%.1f", x)

#theme for all plots
figureargs <- theme_light(base_size = 25)
#+ ylab("log cluster coefficient") + xlab("log degree") +
 # scale_x_continuous( labels=scaleFUN) + scale_y_continuous(labels=scaleFUN)

## tenure per author -------------------------------------

#' The function plots hierarchy data per time range with the additional
#' information, how long a developer is already active in the project.
#' The activity is measured by the time between the the first activity
#' and the current range.
#'
#' @param network.list list of networks regarding the time ranges
#' @param project.data the project data object
#' @param activity a string, which describes the data source
#'                 (\code{c("mail", "commits", "issues")})
#' @param hierarchy.data list of hierarchy data per time range
#' @param plot.path file path to save the plot
#'
hierarchy.plot.with.stay.period <- function(network.list, project.data, activity,
                                            hierarchy.data, plot.path) {
  # compute first activity and add information to the network list
  networks.with.first.activity <- add.vertex.attribute.first.activity(network.list,project.data,
                                            activity.types = activity,
                                            name = "first.activity",
                                            aggregation.level = c("cumulative"))

  # merge hierarchy data and activity data of the networks
  hierarchies.with.period <- lapply(names(hierarchy.data), function(project.name){
    current.hierarchy.data <- hierarchy.data[[project.name]]
    current.network <- networks.with.first.activity[[project.name]]
    current.network.df.vertices <- as_data_frame(current.network, "vertices")

    current.end.date.str <- substr(project.name, 21, 39)
    current.end.date <- strptime(current.end.date.str , "%Y-%m-%d %H:%M:%S")

    hierarchy.with.activity.time.period <- lapply(row.names(current.hierarchy.data),
                                                  function(developer.name){

      if (!is.null(current.network.df.vertices) && nrow(current.network.df.vertices) > 0) {
        complete.first.activity.of.developer <- min(unlist(
          current.network.df.vertices[developer.name, "first.activity"][[1]]), na.rm=TRUE)
        t <- as.difftime(c(as.numeric(current.end.date) - complete.first.activity.of.developer),
                         units = "secs")
        current.hierarchy.data[developer.name, "time.period"] <- as.numeric(t, units="days")

        return(current.hierarchy.data[developer.name,])
      }
    })
    return(hierarchy.with.activity.time.period)
  })

  names(hierarchies.with.period) <- names(hierarchy.data)
  # plot hierarchy with activity data
  plots <- lapply(names(hierarchies.with.period), function(hierarchy.name){
    hierarchy.df <- rbind.fill(hierarchies.with.period[[hierarchy.name]])
    if (!is.null(hierarchy.df)) {
      current.points.plot <- ggplot() +
        geom_point(data = hierarchy.df, aes(jitter(hierarchy.log.deg, factor = 0.5),
                                            size = time.period, colour = time.period,
                                            jitter(hierarchy.log.cc, factor = 0.5))) +
        figureargs +  ylab("log cluster coefficient") + xlab("log degree") +
        scale_x_continuous(labels=scaleFUN) +
        scale_y_continuous(labels=scaleFUN) +
        ggtitle(paste(substr(hierarchy.name,1,10), substr(hierarchy.name,21,30), sep=" to ")) +
        theme(legend.position="right", legend.box = "vertical") +
        scale_size_continuous(name="tenure (days)") +
        scale_color_continuous(name="tenure (days)")
      return(current.points.plot)
    }
  })

  # save plots
  plots <- plyr::compact(plots)
  pdf(file=plot.path, height= 7*ceiling(length(plots)/5), width = 12*5)
  do.call("grid.arrange", c(plots, ncol=5, newpage=TRUE))
  dev.off()
}

## aggregated tenure ------


#' The function computes for each developer for each network in the list,
#' whether he is part of the hierarchy and his tenure in the project.
#'
#' @param network.list list of networks regarding the time ranges
#' @param project.data the project data object
#' @param activity a string, which desrcibes the data source
#'                 (\code{c("mail", "commits", "issues")})
#' @param hierarchy.data list of hierarchy data per time range
#' @return like hierarchy.data with additional parameters \code{part.of.hierarchy}
#' and \code{time.period} for the tenure of the developers
compute.hierarchy.and.tenure <- function(network.list, project.data, activity,
                                            hierarchy.data,
                                         hierarchy.data.with.normalized.criteria) {
  # compute first activity and add information to the network list
  hierarchy.data <- plyr::compact(hierarchy.data)
  networks.with.first.activity <- add.vertex.attribute.first.activity(network.list,
                                                                      project.data,
                                                                      activity.types = activity,
                                                                      name = "first.activity",
                                                                      aggregation.level = c("cumulative"))

  # merge hierarchy data and activity data of the networks
  hierarchies.with.period <- lapply(names(hierarchy.data), function(project.name){
    current.hierarchy.data <- hierarchy.data[[project.name]]
    current.network <- networks.with.first.activity[[project.name]]
    current.network.df.vertices <- as_data_frame(current.network, "vertices")

    current.end.date.str <- substr(project.name, 21, 39)
    current.end.date <- strptime(current.end.date.str , "%Y-%m-%d %H:%M:%S")

    hierarchy.with.activity.time.period <- lapply(row.names(current.hierarchy.data),
                                                  function(developer.name){

      if (!is.null(current.network.df.vertices) && nrow(current.network.df.vertices) > 0) {
        complete.first.activity.of.developer <- min(unlist(
          current.network.df.vertices[developer.name, "first.activity"][[1]]), na.rm=TRUE)
        t <- as.difftime(c(as.numeric(current.end.date) -
                           complete.first.activity.of.developer),
                         units = "secs")
        current.hierarchy.data[developer.name, "time.period"] <- as.numeric(t, units="days")

        return(current.hierarchy.data[developer.name,])
      }
    })
    result = rbind.fill(hierarchy.with.activity.time.period)
    return(result)
  })
  names(hierarchies.with.period) <- names(hierarchy.data)

  # compute hierarchical and non-hierarchical part
  hierarchy.with.both <- lapply(names(hierarchies.with.period), function(project.name) {
    hierarchy.both <- compute.division.into.hierarchical.and.non.hierachical.part(
      hierarchies.with.period[[project.name]],
      number.of.steps = NUMBER.OF.STEPS,
      hierarchy.data.with.normalized.criteria,
      net.name = project.name)
    return(hierarchy.both)
  })
  names(hierarchy.with.both) <- names(hierarchies.with.period)

  return(hierarchy.with.both)
}

#' The function computes the mean tenure for each time range of the hierarchical part
#' and the non-hierarchical part of the network
#'
#' @param hierarchy.with.both result of \code{compute.hierarchy.and.tenure}
#' @return data frame with time range, group and mean tenure for this developer
#' group and time range
compute.mean.tenure <- function(hierarchy.with.both) {
  mean.tenure <- lapply(names(hierarchy.with.both), function(name.time.range){
    hierarchy.both.data <- hierarchy.with.both[[name.time.range]]
    mean.hierarchy <- NA
    mean.periphery <- NA
    logging::logdebug(paste("mean.tenure: name.time.range: ", name.time.range))
    if (is.null(hierarchy.both.data)) {
      return(NULL)
    }

    if (!is.null(hierarchy.both.data$part.of.hierarchy)) {
      mean.hierarchy <- mean(hierarchy.both.data[hierarchy.both.data$part.of.hierarchy,
                                                 "time.period"])
    }
    if (!is.null(hierarchy.both.data[!hierarchy.both.data$part.of.hierarchy,])) {
      mean.periphery <- mean(hierarchy.both.data[!hierarchy.both.data$part.of.hierarchy,
                                                 "time.period"])
    }

    return(data.frame(range = c(name.time.range, name.time.range),
                      group = c("hierarchy", "periphery"),
                      mean.tenure = c(mean.hierarchy, mean.periphery)))
  })

  return(rbind.fill(mean.tenure))
}

#' The function computes the median tenure for each time range of the hierarchical part
#' and the non-hierarchical part of the network
#'
#' @param hierarchy.with.both result of \code{compute.hierarchy.and.tenure}
#' @return data frame with time range, group and mean tenure for this developer
#' group and time range
compute.median.tenure <- function(hierarchy.with.both) {
  mean.tenure <- lapply(names(hierarchy.with.both), function(name.time.range){
    hierarchy.both.data <- hierarchy.with.both[[name.time.range]]
    mean.hierarchy <- NA
    mean.periphery <- NA
    logging::logdebug(paste("mean.tenure: name.time.range: ", name.time.range))
    if (is.null(hierarchy.both.data)) {
      return(NULL)
    }

    if (!is.null(hierarchy.both.data$part.of.hierarchy)) {
      median.hierarchy <- median(hierarchy.both.data[hierarchy.both.data$part.of.hierarchy,
                                                     "time.period"])
    }
    if (!is.null(hierarchy.both.data[!hierarchy.both.data$part.of.hierarchy,])) {
      median.periphery <- median(hierarchy.both.data[!hierarchy.both.data$part.of.hierarchy,
                                                     "time.period"])
    }

    return(data.frame(range = c(name.time.range, name.time.range),
                      group = c("hierarchy", "periphery"),
                      median.tenure = c(median.hierarchy, median.periphery)))
  })

  return(rbind.fill(mean.tenure))
}

#' The function plots the mean tenure for each time range of the hierarchical part
#' and the non-hierarchical part of the network in box plot
#'
#' @param hierarchy.with.both result of \code{compute.mean.tenure}
#' @param plot.output.file path of the output file
plot.hierarchy.and.tenure <- function(mean.tenure.per.hierarchy.group, plot.output.file) {
  aggregated.mean.tenure <- ggplot(mean.tenure.per.hierarchy.group) +
    geom_point(aes(x = range, y = mean.tenure, color = group,
                   fill = group, shape = group), size=2,
               stat = "identity", position = position_dodge(0.8),) +
    scale_color_manual(values=c("cornflowerblue", "darkblue")) +
    scale_fill_manual(values=c("cornflowerblue", "darkblue")) +
    ylab("mean tenure (days)") + xlab("ranges ") +
    figureargs + theme(axis.text.x=element_blank(),
                       axis.text=element_text(size=18),
                       axis.title=element_text(size=24))

  pdf(file=plot.output.file, height= 7, width = 12)
  print(aggregated.mean.tenure)
  dev.off()
}

#' The function plots the mean tenure of the hierarchical part
#' and the non-hierarchical part of the network in a violin plot
#' (Aggregation of the mean tenure values over the complete project history )
#'
#' @param hierarchy.with.both result of \code{compute.mean.tenure}
#' @param plot.output.file path of the output file
plot.hierarchy.and.tenure.violin <- function(mean.tenure.per.hierarchy.group, plot.output.file) {
  aggregated.mean.tenure <- ggplot(mean.tenure.per.hierarchy.group,
                                   aes(x = group, y = mean.tenure)) +
    geom_violin( trim = FALSE, aes(color = group)) +
    scale_color_manual(values=c("cornflowerblue", "darkblue")) +
    #scale_x_discrete(labels=c("FALSE" = "Periphery", "TRUE" = "Hierarchy")) +
    stat_summary(fun.data=mean_sdl, mult=1,  geom="pointrange", color="red") +
    ylab("mean tenure (days)") + xlab("developer group") + figureargs
    theme(axis.text=element_text(size=16),
          axis.title=element_text(size=20), legend.position = "none")

  pdf(file=plot.output.file, height= 6, width = 10)
  print(aggregated.mean.tenure)
  dev.off()
}

#' The function plots the tenure of the hierarchical part
#' and the non-hierarchical part of all networks in one violin plot
#' (Aggregation of the tenure values of the developers of one group into one
#' violin plot)
#'
#' @param hierarchy.with.both result of \code{compute.hierarchy.and.tenure}
#' @param plot.output.file path of the output file
plot.hierarchy.and.tenure.violin.aggregate.all.ranges <- function(hierarchy.with.tenure,
                                                                  plot.output.file) {
  summarized.all.ranges <- rbind.fill(hierarchy.with.tenure)
  aggregated.mean.tenure <- ggplot(data=subset(summarized.all.ranges, !is.na(part.of.hierarchy)),
                                   aes(x = part.of.hierarchy, y = time.period)) +
    geom_violin(trim = TRUE, aes(color = part.of.hierarchy)) +
    scale_color_manual(values=c("cornflowerblue", "darkblue")) +
    scale_x_discrete(labels=c("FALSE" = "Periphery", "TRUE" = "Hierarchy")) +
    stat_summary(fun.data=mean_sdl, mult=1,  geom="pointrange", color="red") +
    ylab("tenure (days)") + xlab("developer group") + figureargs +
    theme(axis.text=element_text(size=18),
          axis.title=element_text(size=24),
          legend.position = "none")

  pdf(file=plot.output.file, height= 7, width = 12)
  print(aggregated.mean.tenure)
  dev.off()
}

#' The function plots the tenure of the hierarchical part
#' and the non-hierarchical part of all networks in one violin plot for
#' each range
#' (Aggregation of the tenure values of the developers of one group into one
#' violin plot for one range)
#'
#' @param hierarchy.with.both result of \code{compute.hierarchy.and.tenure}
#' @param plot.output.file path of the output file
plot.hierarchy.and.tenure.violin.per.range <- function(hierarchy.with.tenure, plot.output.file) {
  plots <- lapply(names(hierarchy.with.tenure), function(range.name){
    hier.with.ten <- hierarchy.with.tenure[[range.name]]

    if (is.null(hier.with.ten)) {
      return(NULL)
    }

    current.plot <- ggplot(hier.with.ten, aes(x = part.of.hierarchy, y = time.period)) +
      geom_violin(trim = FALSE, aes(color = part.of.hierarchy)) +
      scale_color_manual(values=c("cornflowerblue", "darkblue"),
                         labels=c("periphery", "hierarchy")) +
      scale_x_discrete(labels=c("FALSE" = "Periphery", "TRUE" = "Hierarchy")) +
      stat_summary(fun.data=mean_sdl, mult=1,  geom="pointrange", color="red") +
      ylab("mean tenure (days)") + xlab("developer group") + figureargs +
      theme(axis.text=element_text(size=16),
            axis.title=element_text(size=20), legend.position = "none") +
      ggtitle(paste(CASESTUDY, ": ", range.name))

    return(current.plot)
  })

  plots <- plyr::compact(plots)

  # save plots
  pdf(file=plot.output.file, height= 7*ceiling(length(plots)/5), width = 12*5)
  do.call("grid.arrange", c(plots, ncol=5, newpage=TRUE))
  dev.off()
}

#' The function plots the tenure of the hierarchical part
#' and the non-hierarchical part of all networks in one plot with an extra
#' violin for each range
#'
#' @param hierarchy.with.both result of \code{compute.hierarchy.and.tenure}
#' @param plot.output.file path of the output file
plot.hierarchy.and.tenure.many.small.violin.in.one <- function(hierarchy.with.tenure,
                                                               plot.output.file) {
  number.ranges <- length(hierarchy.with.tenure)
  summarized.all.ranges <- rbind.fill(hierarchy.with.tenure)

  aggregated.mean.tenure <- ggplot() +
    geom_violin(data=subset(summarized.all.ranges, !is.na(part.of.hierarchy)),
                aes(x = range, y = time.period, color = part.of.hierarchy),
                trim = TRUE, #, fill = part.of.hierarchy
                position = "dodge", adjust = .5, scale = "width",
                draw_quantiles = c(0.5)) +
    scale_color_manual(values=c("cornflowerblue", "darkblue"),
                       labels=c("Periphery", "Hierarchy"),
                       name = "Developer Group") +
    scale_fill_manual(values=c("cornflowerblue", "darkblue"),
                      labels=c("Periphery", "Hierarchy"),
                      name = "Developer Group") +
    ylab("tenure (days)") + xlab("time") + figureargs +
    theme(axis.text=element_text(size=16),
          axis.title=element_text(size=20), axis.text.x=element_text(angle=90))

  pdf(file=plot.output.file, height= 14, width = 2*number.ranges)
  print(aggregated.mean.tenure)
  dev.off()
}

plot.hierarchy.and.tenure.many.small.violin.in.one.with.bullet <- function(hierarchy.with.tenure, median.values,
                                                               plot.output.file) {
  number.ranges <- length(hierarchy.with.tenure)
  summarized.all.ranges <- rbind.fill(hierarchy.with.tenure)

  aggregated.mean.tenure <- ggplot() +
    geom_violin(data=subset(summarized.all.ranges, !is.na(part.of.hierarchy)),
                aes(x = range, y = time.period, color = part.of.hierarchy),
                trim = TRUE, #, fill = part.of.hierarchy
                position = "dodge", adjust = .5, scale = "width",
                draw_quantiles = c(0.5)) +
    geom_violin(data=subset(summarized.all.ranges, !is.na(part.of.hierarchy)),
                aes(x = range, y = time.period, color = part.of.hierarchy),
                trim = TRUE, #, fill = part.of.hierarchy
                position = "dodge", adjust = .5, scale = "width", size=2,
                draw_quantiles = c(0.25, 0.7)) +
    scale_color_manual(values=c("cornflowerblue", "darkred", "red" , "darkblue"), #"skyblue1", "navy"
                       labels=c("Periphery", "Median of Hierarchy", "Median of Periphery", "Hierarchy"),
                       name = "Developer Group") +
    scale_fill_manual(values=c("cornflowerblue", "darkblue", "red", "orange"),
                      labels=c("Periphery", "Median of Hierarchy", "Median of Periphery", "Hierarchy"),
                      name = "Developer Group") +
    ylab("tenure (days)") + xlab("time") + figureargs +
    theme(axis.text=element_text(size=16),
          axis.title=element_text(size=20), axis.text.x=element_text(angle=90)) +
    geom_point(data = median.values, aes(x= range, y= median.tenure, color = group, size =4), na.rm = TRUE)

  pdf(file=plot.output.file, height= 14, width = 2*number.ranges)
  print(aggregated.mean.tenure)
  dev.off()
}

#' The function compute statistical tests.
#'
#' @param hierarchy.with.both result of \code{compute.hierarchy.and.tenure}
#' @param plot.output.file path of the output file
compute.statistical.tests.tenure <- function(hierarchy.with.tenure, test.output.file) {
  summarized.all.ranges <- rbind.fill(hierarchy.with.tenure)
  t.test.tenure <- t.test(x = summarized.all.ranges[summarized.all.ranges$part.of.hierarchy,
                                                    "time.period"],
                          y = summarized.all.ranges[!summarized.all.ranges$part.of.hierarchy,
                                                    "time.period"])

  mann.whitney.u.test <- wilcox.test(x = summarized.all.ranges[summarized.all.ranges$part.of.hierarchy,
                                                               "time.period"],
                                     y = summarized.all.ranges[!summarized.all.ranges$part.of.hierarchy,
                                                               "time.period"])
  effect.size.wilcoxon <- rcompanion::wilcoxonR(x = summarized.all.ranges$time.period,
                                                g = summarized.all.ranges$part.of.hierarchy)

  eff.cliff.delta <- effsize::cliff.delta(d = summarized.all.ranges[summarized.all.ranges$part.of.hierarchy,
                                                                    "time.period"],
                                          f = summarized.all.ranges[!summarized.all.ranges$part.of.hierarchy,
                                                                    "time.period"])
  median.hierachy = median(summarized.all.ranges[summarized.all.ranges$part.of.hierarchy,
                                                 "time.period"], na.rm = TRUE)
  median.periphery = median(summarized.all.ranges[!summarized.all.ranges$part.of.hierarchy,
                                                  "time.period"], na.rm = TRUE)

  statistic.tests <- data.frame("t.test-p"=t.test.tenure$p.value,
                                "t-test-smaller-0.001" =
                                  (t.test.tenure$p.value < 0.001),
                                "mann.whitney.u.test-p"= mann.whitney.u.test$p.value,
                                "mann.whitney.u.test-threshold-smaller-0.001" =
                                  (mann.whitney.u.test$p.value < 0.001),
                                "effect.size"= effect.size.wilcoxon,
                                "effect.size-smaller-0.5" =
                                  (effect.size.wilcoxon < 0.5),
                                "cliff.delta"= eff.cliff.delta$estimate,
                                "number.hierarchy.devs"= length(
                                  summarized.all.ranges[summarized.all.ranges$part.of.hierarchy,
                                                        "time.period"]),
                                "number.periphery.devs"= length(
                                  summarized.all.ranges[!summarized.all.ranges$part.of.hierarchy,
                                                        "time.period"]),
                                "difference medians" = median.hierachy - median.periphery,
                                "difference.medians.normalized" = (median.hierachy - median.periphery)/median.hierachy
  )
  pdf(file=test.output.file, height= 6, width = 25)
  grid.table(statistic.tests)
  dev.off()
}


## number edited files per author --------------------------------------------------

#' The function plots the hierarchy data with point size regarding the number of
#' artifacts a developer edits. The list of plots is saved in the file with the
#' file path \code{plot.path}.
#'
#' @param network.list list of networks per time range
#' @param project.data project data object (necessary to compute the number of
#'                     edited artifacts)
#' @param activity selected artifact type (not yet used)
#' @param hierarchy.data list of hierarchy data with a data set for each time
#'                       range
#' @param plot.path file path for the result file with the hierarchy plots
compute.artifact.count <- function(network.list, project.data, activity, hierarchy.data) {

  # count artifacts per developer
  networks <- add.vertex.attribute.artifact.count(network.list,project.data,
                                                  # activity.types = activity,
                                                  name = "artifact.count",
                                                  aggregation.level = c("range"))

  # merge hierarchy and artifact count data
  #TODO: nur commit counts
  hierarchies.with.artifact.count <- lapply(names(hierarchy.data), function(project.name) {
    current.hierarchy.data <- hierarchy.data[[project.name]]
    current.network <- networks[[project.name]]
    current.network.df.vertices <- as_data_frame(current.network, "vertices")

    hierarchy.with.artifact.count <- lapply(row.names(current.hierarchy.data),
                                            function(developer.name){
      if (!is.null(current.network.df.vertices) && nrow(current.network.df.vertices) > 0) {
        current.hierarchy.data[developer.name, "artifact.count"] <-
        current.network.df.vertices[developer.name,"artifact.count"]
        return(current.hierarchy.data[developer.name,])
      } else {
        logging::loginfo(paste("hierarchy.plot.with.artifact.count: ",
                               "No artifact.count for developer ",
                               developer.name, "."))
        return(NULL)
      }
    })
    return(hierarchy.with.artifact.count)
  })

  names(hierarchies.with.artifact.count) <- names(hierarchy.data)

  hierarchies.with.artifact.count <- plyr::compact(hierarchies.with.artifact.count)
  return(hierarchies.with.artifact.count)
}


#' @param hierarchies.with.artifact.count result of \code{compute.artifact.count}
plot.hierarchy.with.artifact.count <- function(hierarchies.with.artifact.count, plot.path) {

  # plot hierarchy with artifact count data
  plots <- lapply(names(hierarchies.with.artifact.count), function(hierarchy.name){
    hierarchy <- hierarchies.with.artifact.count[[hierarchy.name]]
    if (!is.null(hierarchy) && length(hierarchy) > 0) {
      hierarchy.df <- rbind.fill(hierarchy)
      current.points.plot <- ggplot() +
      geom_point(data = hierarchy.df, aes(jitter(hierarchy.log.deg, factor = 0.5),
                                          size = artifact.count, colour=artifact.count,
                                          jitter(hierarchy.log.cc, factor = 0.5))) +
      figureargs +
        ggtitle(paste(substr(hierarchy.name,1,10), substr(hierarchy.name,21,30), sep=" to ")) +
        ylab("log cluster coefficient") + xlab("log degree") +
        scale_x_continuous( labels=scaleFUN) +
        scale_y_continuous(labels=scaleFUN) +
        theme(legend.position="right", legend.box = "vertical") +
        scale_size_continuous(name="number edited files") +
        scale_color_continuous(name="number edited files")
      return(current.points.plot)
    }
  })

  plots <- plyr::compact(plots)

  # save plots
  pdf(file=plot.path, height= 7*ceiling(length(plots)/5), width = 12*5)
  do.call("grid.arrange", c(plots, ncol=5, newpage=TRUE))
  dev.off()
}

## Aggregated edited files -----------------------------

compute.hierarchy.and.artifact.count <- function(hierarchies.with.artifact.count,
                                                 hierarchy.data.with.normalized.criteria) {
  logging::loginfo("Start: compute.hierarchy.and.artifact.count")
  hierarchies.with.artifact.count <- plyr::compact(hierarchies.with.artifact.count)
  # compute hierarchical and non-hierarchical part
  hierarchy.with.both <- lapply(names(hierarchies.with.artifact.count), function(project.name) {
    hierarchy.both <- compute.division.into.hierarchical.and.non.hierachical.part(
      hierarchies.with.artifact.count[[project.name]],
      number.of.steps = NUMBER.OF.STEPS,
      hierarchy.data.with.normalized.criteria,
      net.name = project.name)
    return(hierarchy.both)
  })
  names(hierarchy.with.both) <- names(hierarchies.with.artifact.count)
  return(hierarchy.with.both)
}


#' The function computes the mean edited files for each time range of the hierarchical part
#' and the non-hierarchical part of the network
#'
#' @param hierarchy.with.both result of \code{comput.hierarchy.and.artifact.count}
#' @return data frame with time range, group and mean edited files for this developer
#' group and time range
compute.mean.edited.files <- function(hierarchy.with.both) {
  mean.edited.files <- lapply(names(hierarchy.with.both), function(name.time.range){
    hierarchy.both.data <- hierarchy.with.both[[name.time.range]]
    mean.hierarchy <- NA
    mean.periphery <- NA
    logging::logdebug(paste("mean.edited.files: name.time.range: ", name.time.range))
    if (is.null(hierarchy.both.data)) {
      return(NULL)
    }

    if (!is.null(hierarchy.both.data$part.of.hierarchy)) {
      mean.hierarchy <- mean(hierarchy.both.data[hierarchy.both.data$part.of.hierarchy,
                                                 "artifact.count"])
    }
    if (!is.null(hierarchy.both.data[!hierarchy.both.data$part.of.hierarchy,])) {
      mean.periphery <- mean(hierarchy.both.data[!hierarchy.both.data$part.of.hierarchy,
                                                 "artifact.count"])
    }

    return(data.frame(range = c(name.time.range, name.time.range),
                      group = c("hierarchy", "periphery"),
                      mean.artifact.count = c(mean.hierarchy, mean.periphery)))
  })

  return(rbind.fill(mean.edited.files))
}

#' The function compute statistical tests.
#'
#' @param hierarchy.with.both result of \code{compute.hierarchy.and.edited.files}
#' @param plot.output.file path of the output file
compute.statistical.tests.edited.files <- function(hierarchy.with.edited.files, test.output.file) {
  summarized.all.ranges <- rbind.fill(hierarchy.with.edited.files)
  t.test.edited.files <- t.test(x = summarized.all.ranges[summarized.all.ranges$part.of.hierarchy,
                                                          "artifact.count"],
                                y = summarized.all.ranges[!summarized.all.ranges$part.of.hierarchy,
                                                         "artifact.count"])

  mann.whitney.u.test <- wilcox.test(x = summarized.all.ranges[summarized.all.ranges$part.of.hierarchy,
                                                               "artifact.count"],
                                     y = summarized.all.ranges[!summarized.all.ranges$part.of.hierarchy,
                                                               "artifact.count"])
  effect.size.wilcoxon <- rcompanion::wilcoxonR(x = summarized.all.ranges$artifact.count,
                                                g = summarized.all.ranges$part.of.hierarchy)

  eff.cliff.delta <- effsize::cliff.delta(d = summarized.all.ranges[summarized.all.ranges$part.of.hierarchy,
                                                                    "artifact.count"],
                                          f= summarized.all.ranges[!summarized.all.ranges$part.of.hierarchy,
                                                                   "artifact.count"])
  median.hierachy = median(summarized.all.ranges[summarized.all.ranges$part.of.hierarchy,
                                                 "artifact.count"], na.rm = TRUE)
  median.periphery = median(summarized.all.ranges[!summarized.all.ranges$part.of.hierarchy,
                                                  "artifact.count"], na.rm = TRUE)

  statistic.tests <- data.frame("t.test-p"=t.test.edited.files$p.value,
                                "t-test-smaller-0.001" =
                                  (t.test.edited.files$p.value < 0.001),
                                "mann.whitney.u.test-p"= mann.whitney.u.test$p.value,
                                "mann.whitney.u.test-threshold-smaller-0.001" =
                                  (mann.whitney.u.test$p.value < 0.001),
                                "effect.size"= effect.size.wilcoxon,
                                "effect.size-smaller-0.5" =
                                  (effect.size.wilcoxon < 0.5),
                                "cliff.delta"= eff.cliff.delta$estimate,
                                "number.hierarchy.devs"= length(
                                  summarized.all.ranges[summarized.all.ranges$part.of.hierarchy,
                                                        "time.period"]),
                                "number.periphery.devs"= length(
                                  summarized.all.ranges[!summarized.all.ranges$part.of.hierarchy,
                                                        "time.period"]),
                                "difference medians" = median.hierachy - median.periphery,
                                "difference.medians.normalized" = (median.hierachy - median.periphery)/median.hierachy
  )
  pdf(file=test.output.file, height= 6, width = 25)
  grid.table(statistic.tests)
  dev.off()
}

#' The function plots the mean edited files for each time range of the hierarchical part
#' and the non-hierarchical part of the network in box plot
#'
#' @param hierarchy.with.both result of \code{compute.mean.edited.files}
#' @param plot.output.file path of the output file
plot.hierarchy.and.edited.files <- function(mean.edited.files.per.hierarchy.group, plot.output.file) {
  aggregated.mean.edited.files <- ggplot(mean.edited.files.per.hierarchy.group) +
    geom_point(aes(x = range, y = mean.artifact.count, color = group,
                   fill = group, shape = group), size=2,
               stat = "identity", position = position_dodge(0.8),) +
    scale_color_manual(values=c("cornflowerblue", "darkblue")) +
    scale_fill_manual(values=c("cornflowerblue", "darkblue")) +
    ylab("mean number of edited files") + xlab("ranges ") +
    figureargs + theme(axis.text.x=element_blank(),
                       axis.text=element_text(size=18),
                       axis.title=element_text(size=24))

  pdf(file=plot.output.file, height= 7, width = 12)
  print(aggregated.mean.edited.files)
  dev.off()
}

#' The function plots the edited files of the hierarchical part
#' and the non-hierarchical part of all networks in one violin plot
#' (Aggregation of the numbers of edited files of the developers of one group into one
#' violin plot)
#'
#' @param hierarchy.with.both result of \code{compute.hierarchy.and.edited.files}
#' @param plot.output.file path of the output file
plot.hierarchy.and.edited.files.violin.aggregate.all.ranges <- function(hierarchy.with.edited.files,
                                                                  plot.output.file) {
  summarized.all.ranges <- rbind.fill(hierarchy.with.edited.files)
  aggregated.mean.edited.files <- ggplot(data=subset(summarized.all.ranges, !is.na(part.of.hierarchy)),
                                         aes(x = part.of.hierarchy, y = artifact.count)) +
    geom_violin( trim = TRUE, aes(color = part.of.hierarchy)) +
    scale_color_manual(values=c("cornflowerblue", "darkblue")) +
    scale_x_discrete(labels=c("FALSE" = "Periphery", "TRUE" = "Hierarchy")) +
    stat_summary(fun.data=mean_sdl, mult=1,  geom="pointrange", color="red") +
    ylab("number of edited files") + xlab("developer group") + figureargs +
    theme(axis.text=element_text(size=18),
          axis.title=element_text(size=24), legend.position = "none")

  pdf(file=plot.output.file, height= 7, width = 12)
  print(aggregated.mean.edited.files)
  dev.off()
}

## Compute maximum bin size for movement plots --------------------------------

compute.maximal.bin.size.for.movement <- function(hierarchy.data, selected.developers) {

  # add time step  and normalized values to hierarchy data
  for (i in 1:length(hierarchy.data)){
    if (!is.null(hierarchy.data[[i]]) && !is.null(nrow(hierarchy.data[[i]]))) {
      hierarchy.data[[i]]$time.step = i
      max.log.deg <- max(hierarchy.data[[i]]$hierarchy.log.deg)
      min.log.cc <- min(hierarchy.data[[i]]$hierarchy.log.cc)
      hierarchy.data[[i]]$log.norm.deg <- hierarchy.data[[i]]$hierarchy.log.deg / max.log.deg
      hierarchy.data[[i]]$log.norm.cc <- - hierarchy.data[[i]]$hierarchy.log.cc / min.log.cc
    }
  }

  hierarchy.data.of.selected.developer <- lapply(selected.developers,
                                                 function(developer.name){
    hierarchy.data.of.current.developer <- lapply(hierarchy.data,
                                                  function(hierarchy){
      if (developer.name %in% row.names(hierarchy)) {
        return(hierarchy[developer.name, ])
      }
    })
    return(data.frame("developer.name" = developer.name, rbind.fill(rbind.fill(hierarchy.data.of.current.developer))))
  })

  max.bin.per.developer <- lapply(1:length(hierarchy.data.of.selected.developer),
                                  function(current.developer) {
    hierarchy <- hierarchy.data.of.selected.developer[[current.developer]]
    hierarchy$bin.cc <- .bincode(hierarchy$log.norm.cc, breaks = seq(-1.1,0.125,0.125))
    hierarchy$bin.deg <- .bincode(hierarchy$log.norm.deg, breaks = seq(-0.125,1.1,0.125))
    df2 <- count(hierarchy, c('bin.deg','bin.cc'))

    return(data.frame("value" = max(df2$freq), "developer.name"=hierarchy$developer.name[1]))
  })
  max.df <- rbind.fill(max.bin.per.developer)
  return(max(max.df$value))
}

## movement in the hierarchy of selected developers ( 2 plots but 1 file) ----------------

#' Similar to the function \code{plot.trajectory.in.hierarchy}, but using heap
#' maps instead of trajectory and ignoring the time.
#'
#' @param hierarchy.data list of hierarchy.data with one data set for each time range
#' @param selected.developer.names names of selected developers, who are analysed
#' @param plot.path file path, where the result plot is saved
plot.movement.in.hierarchy <- function(hierarchy.data, selected.developer.names,
                                       plot.path, max.bins) {
  # add time step to hierarchy data
  for (i in 1:length(hierarchy.data)){
    if (!is.null(hierarchy.data[[i]]) && !is.null(nrow(hierarchy.data[[i]]))) {
      hierarchy.data[[i]]$time.step = i
      max.log.deg <- max(hierarchy.data[[i]]$hierarchy.log.deg)
      min.log.cc <- min(hierarchy.data[[i]]$hierarchy.log.cc)
      hierarchy.data[[i]]$log.norm.deg <- hierarchy.data[[i]]$hierarchy.log.deg / max.log.deg
      hierarchy.data[[i]]$log.norm.cc <- - hierarchy.data[[i]]$hierarchy.log.cc / min.log.cc
    }
  }

  # get hierarchy data for the selected developers
  hierarchy.data.of.selected.developer <- lapply(selected.developer.names,
                                                 function(developer.name){
    hierarchy.data.of.current.developer <- lapply(hierarchy.data,
                                                  function(hierarchy){
      if (developer.name %in% row.names(hierarchy)) {
        return(hierarchy[developer.name, ])
      }
    })
    return(data.frame("developer.name"= developer.name, rbind.fill(rbind.fill(hierarchy.data.of.current.developer))))
  })

  # plot the hierarchy data of the selected developer in a heap map
  if (length(selected.developer.names) == 1) {
    hierarchy.df <- rbind.fill(hierarchy.data.of.selected.developer)
    heap.map.plot <- ggplot(data = hierarchy.df, aes(log.norm.deg, log.norm.cc)) +
      stat_bin2d(aes(fill = stat(count)), binwidth = c(0.125,0.125)) +
      xlab("standardized degree") + ylab("standardized cluster coefficient") +
      scale_fill_gradientn(colours = c("#FFFF66", "#FF6600", "#CC0000"), limits=c(1, max.bins)) +
      scale_x_continuous(breaks = c(0, 0.5, 1), limits=c(-0.126, 1.126)) +
      scale_y_continuous(breaks = c(-1, -0.5, 0), limits = c(-1.126, 0.126)) + figureargs
  } else {
    hierarchy.df <- rbind.fill(hierarchy.data.of.selected.developer)
    heap.map.plot <- ggplot(data = hierarchy.df, aes(log.norm.deg, log.norm.cc)) +
      stat_bin2d(aes(fill = stat(count)), binwidth = c(0.125,0.125)) +
      xlab("standardized degree") + ylab("standardized cluster coefficient") + xlim(0, 1.125) +
      scale_fill_gradient(low = "#33CC00", high = "#000099") + figureargs
  }

  # plot the trajectory
  if (length(selected.developer.names) == 1) {
    # only one developer (add time information to the plot)
    hierarchy.df <- rbind.fill(hierarchy.data.of.selected.developer)
    trajectory.plot <- ggplot() +
      geom_path(data = hierarchy.df, aes(log.norm.deg, log.norm.cc,
                                         group=developer.name,
                                         colour=time.step), lwd=1.5, lineend = "round") +
      xlab("standardized degree") +
      ylab("standardized cluster coefficient") + #ggtitle(selected.developer.names) +
      #xlim(-0.126, 1.126) + ylim(-1.126, 0.126) +
      scale_colour_gradient(limits=c(1, length(hierarchy.data)), name="time steps") + figureargs +
      scale_x_continuous(breaks = c(0, 0.5, 1), limits=c(-0.126, 1.126)) +
      scale_y_continuous(breaks = c(-1, -0.5, 0), limits = c(-1.126, 0.126)) +
      # Start point
      geom_point(data = hierarchy.df[1,],
                 shape= 17, size=5, colour= "royalblue4" ,
                 aes(log.norm.deg, log.norm.cc)) +
      # End point
      geom_point(data = hierarchy.df[nrow(hierarchy.df),],
                 shape= 15, size=5, colour= "steelblue2",
                 aes(log.norm.deg, log.norm.cc))

  } else {
    # more than one developer
    hierarchy.df <- rbind.fill(hierarchy.data.of.selected.developer)
    trajectory.plot <- ggplot() +
      geom_path(data = hierarchy.df, aes(log.norm.deg, log.norm.cc,
                                         group=developer.name,
                                         colour=developer.name),
                arrow = arrow()) + xlab("log.degree") + ylab("log.clus.coef") + figureargs
  }

  # save heap map plot
  # pdf(file=plot.path, height= 6, width = 10)
  # print(c(trajectory.plot, heap.map.plot))
  # dev.off()

  plots <- list()
  plots[[1]] <- trajectory.plot
  plots[[2]] <- heap.map.plot
  pdf(file=plot.path, height= 6, width = 9*2)
  do.call("grid.arrange", c(plots, ncol=2, newpage=TRUE))
  dev.off()
}


## combination of heap map and trajectory movement in the hierarchy of selected developers ---------

#' Similar to the function \code{plot.trajectory.in.hierarchy}, but using heap
#' maps instead of trajectory and ignoring the time.
#'
#' @param hierarchy.data list of hierarchy.data with one data set for each time range
#' @param selected.developer.names names of selected developers, who are analysed
#' @param plot.path file path, where the result plot is saved
plot.movement.in.hierarchy.combination <- function(hierarchy.data, selected.developer.names,
                                       plot.path, max.bins = 16) {
  # add time step to hierarchy data and normalize (?) data
  for (i in 1:length(hierarchy.data)){
    if (!is.null(hierarchy.data[[i]]) && !is.null(nrow(hierarchy.data[[i]]))) {
      hierarchy.data[[i]]$time.step = i
      max.log.deg <- max(hierarchy.data[[i]]$hierarchy.log.deg)
      min.log.cc <- min(hierarchy.data[[i]]$hierarchy.log.cc)
      hierarchy.data[[i]]$log.norm.deg <- hierarchy.data[[i]]$hierarchy.log.deg / max.log.deg
      hierarchy.data[[i]]$log.norm.cc <- - hierarchy.data[[i]]$hierarchy.log.cc / min.log.cc
    }
  }

  # get hierarchy data for the selected developers
  hierarchy.data.of.selected.developer <- lapply(selected.developer.names,
                                                 function(developer.name){
    hierarchy.data.of.current.developer <- lapply(hierarchy.data,
                                                  function(hierarchy){
      if (developer.name %in% row.names(hierarchy)) {
        return(hierarchy[developer.name, ])
      }
    })
    return(data.frame("developer.name"= developer.name, rbind.fill(rbind.fill(hierarchy.data.of.current.developer))))
  })

  # plot the hierarchy data of the selected developer in a heap map and a trajectory
  if (length(selected.developer.names) == 1) {
    # only one developer (add time information to the plot)
    hierarchy.df <- rbind.fill(hierarchy.data.of.selected.developer)

    trajectory.plot <- ggplot(data = hierarchy.df, aes(log.norm.deg, log.norm.cc))  +
      #heap map
      stat_bin2d(aes(fill = stat(count)), binwidth = c(0.125,0.125)) +
      scale_fill_gradientn(colours = c("#FFFF66", "#FF6600", "#CC0000"), limits=c(1,max.bins)) +
      # trajectory
      geom_path(data = hierarchy.df, aes(log.norm.deg, log.norm.cc,
                                         group=developer.name,
                                         colour=time.step), lwd=1.5, lineend = "round") +
      xlab("standardized degree") +
      ylab("standardized cluster coefficient") + #ggtitle(selected.developer.names) +
      scale_x_continuous(breaks = c(0, 0.5, 1), limits=c(-0.126, 1.126)) +
      scale_y_continuous(breaks = c(-1, -0.5, 0), limits = c(-1.126, 0.126)) +
      scale_colour_gradient(limits=c(1, length(hierarchy.data)), name = "time steps") +
      figureargs +
      # Start point
      geom_point(data = hierarchy.df[1,],
                 shape= 17, size=5, colour="royalblue4",
                 aes(log.norm.deg, log.norm.cc)) +
      # End point
      geom_point(data = hierarchy.df[nrow(hierarchy.df),],
                 shape= 15, size=5, colour="steelblue2",
                 aes(log.norm.deg, log.norm.cc))
  } else {
    # more than one developer
    hierarchy.df <- rbind.fill(hierarchy.data.of.selected.developer)
    # heap map
    trajectory.plot <- ggplot(data = hierarchy.df, aes(log.norm.deg, log.norm.cc)) +
      stat_bin2d(aes(fill = stat(count)), binwidth = c(0.1,0.1)) +
      xlab("log.degree") + ylab("log.clus.coef") + xlim(0, 6) +
      scale_fill_gradient(low = "#33CC00", high = "#000099") + figureargs

    # trajectory
    trajectory.plot <- trajectory.plot +
      geom_path(data = hierarchy.df, aes(log.norm.deg, log.norm.cc,
                                         group=developer.name,
                                         colour=developer.name),
                arrow = arrow()) + xlab("log.degree") + ylab("log.clus.coef") + figureargs

  }

  # save plot
  pdf(file=plot.path, height= 6, width = 10)
  print(trajectory.plot)
  dev.off()
}



## movement in the hierarchy of selected developers ----------------------------

#' This function compute the data for trajectory and plots the trajectories of
#' selected developers. Aggregate the movement information of selected developers
#' in one plot using trajectories.
#'
#' @param hierarchy.data list of hierarchy data; one data set per time range
#' @param selected.developer.names names of the selected developers, for which
#'                                 trajectories are computed
#' @param plot.path file path, where the result plot with the trajectory is saved
plot.trajectory.in.hierarchy <- function(hierarchy.data,
                                         selected.developer.names, plot.path) {
  # add time step to hierarchy.data
  for (i in 1:length(hierarchy.data)){
    hierarchy.data[[i]]$time.step = i
    max.log.deg <- max(hierarchy.data[[i]]$hierarchy.log.deg)
    min.log.cc <- min(hierarchy.data[[i]]$hierarchy.log.cc)
    hierarchy.data[[i]]$log.norm.deg <- hierarchy.data[[i]]$hierarchy.log.deg / max.log.deg
    hierarchy.data[[i]]$log.norm.cc <- - hierarchy.data[[i]]$hierarchy.log.cc / min.log.cc
  }

  # get hierarchy data for the selected developers
  hierarchy.data.of.selected.developer <- lapply(selected.developer.names,
                                                 function(developer.name){
    hierarchy.data.of.current.developer <- lapply(hierarchy.data, function(hierarchy){
      if (developer.name %in% row.names(hierarchy)) {
        return(hierarchy[developer.name, ])
      }
    })
    return(data.frame("developer.name"= developer.name, rbind.fill(rbind.fill(hierarchy.data.of.current.developer))))
  })

  # plot the trajectory
  if (length(selected.developer.names) == 1) {
    # only one developer (add time information to the plot)
    hierarchy.df <- rbind.fill(hierarchy.data.of.selected.developer)
    trajectory.plot <- ggplot() +
      geom_path(data = hierarchy.df, aes(log.norm.deg, log.norm.cc,
                                         group=developer.name,
                                         colour=time.step)) +
      xlab("log.degree") + ylab("log.clus.coef") + #ggtitle(selected.developer.names) +
      xlim(-0.126, 1.126) + ylim(-1.126, 0.126) +
      scale_colour_gradient(limits=c(1, length(hierarchy.data))) + figureargs
  } else {
    # more than one developer
    hierarchy.df <- rbind.fill(hierarchy.data.of.selected.developer)
    trajectory.plot <- ggplot() +
      geom_path(data = hierarchy.df, aes(log.norm.deg, log.norm.cc,
                                         group=developer.name,
                                         colour=developer.name),
                arrow = arrow()) + xlab("log.degree") + ylab("log.clus.coef") + figureargs
  }

  # save plot
  pdf(file=plot.path, height= 6, width = 10)
    print(trajectory.plot)
  dev.off()
}

## movement in the hierarchy of selected developers (heap maps) ----------------

#' Similar to the function \code{plot.trajectory.in.hierarchy}, but using heap
#' maps instead of trajectory and ignoring the time.
#'
#' @param hierarchy.data list of hierarchy.data with one data set for each time range
#' @param selected.developer.names names of selected developers, who are analysed
#' @param plot.path file path, where the result plot is saved
plot.heap.map.in.hierarchy <- function(hierarchy.data, selected.developer.names,
                                       plot.path, max.bins = 16) {
  # add time step to hierarchy data
  for (i in 1:length(hierarchy.data)){
    hierarchy.data[[i]]$time.step = i

  }

  # get hierarchy data for the selected developers
  hierarchy.data.of.selected.developer <- lapply(selected.developer.names,
                                                 function(developer.name){
    hierarchy.data.of.current.developer <- lapply(hierarchy.data,
                                                  function(hierarchy){
      if (developer.name %in% row.names(hierarchy)) {
        return(hierarchy[developer.name, ])
      }
    })
    return(data.frame("developer.name"= developer.name, rbind.fill(rbind.fill(hierarchy.data.of.current.developer))))
  })

  # plot the hierarchy data of the selected developer in a heap map
  if (length(selected.developer.names) == 1) {
    hierarchy.df <- rbind.fill(hierarchy.data.of.selected.developer)
    heap.map.plot <- ggplot(data = hierarchy.df, aes(hierarchy.log.deg, hierarchy.log.cc)) +
      stat_bin2d(aes(fill = stat(count)), binwidth = c(0.25,0.25)) +
      xlab("log.degree") + ylab("log.clus.coef") + #ggtitle(selected.developer.names) +
      scale_fill_gradientn(colours = c("#FFFF66", "#FF6600", "#CC0000"), limits=c(1,max.bins)) +
      # TODO: change limits
      xlim(0, 6) + ylim(-3, 0.25) + figureargs
  } else {
    hierarchy.df <- rbind.fill(hierarchy.data.of.selected.developer)
    heap.map.plot <- ggplot(data = hierarchy.df, aes(hierarchy.log.deg, hierarchy.log.cc)) +
      stat_bin2d(aes(fill = stat(count)), binwidth = c(0.1,0.1)) +
      xlab("log.degree") + ylab("log.clus.coef") + xlim(0, 6) +
      scale_fill_gradient(low = "#33CC00", high = "#000099") + figureargs
  }

  # save heap map plot
  pdf(file=plot.path, height= 6, width = 10)
  print(heap.map.plot)
  dev.off()
  print(heap.map.plot)
}

## mark selected developers in the hierarchy history --------------------------

#' This function plots the hierarchy data and marks selected developers
#' to track the position in the hierarchy of the selected developers over time.
#'
#' @param hierarchy.data list of hierarchy data with one data set per time
#'                       range
#' @param selected.developer.names vector of developer names, which are marked
#'                                 in the the known hierarchy plots
#' @param plot.path file path, where the result plot is saved
plot.with.marked.selected.developers <- function(hierarchy.data,
                                                 selected.developer.names, plot.path) {
  # add developer names to the hierarchy data and more the selected ones
  hierarchy.with.developer.names <- lapply(hierarchy.data, function(hierarchy){
    hierarchy$developer.names <- row.names(hierarchy)

    hierarchy$selected.developer <- FALSE
    hierarchy[hierarchy$developer.names %in% selected.developer.names,
              "selected.developer"] <- TRUE
    return(hierarchy)
  })

  # plot the hierarchy and mark the selected developers with different colors
  plots <- lapply(names(hierarchy.with.developer.names), function(hierarchy.name){
    hierarchy <- hierarchy.with.developer.names[[hierarchy.name]]
    hierarchy.df <- rbind.fill(hierarchy)
    current.points.plot <- ggplot() +
      geom_point(data = hierarchy.df, aes(jitter(hierarchy.log.deg, factor = 0.5),
                                          jitter(hierarchy.log.cc, factor = 0.5))) +
      # theme(legend.position="none") +
      figureargs +
      ggtitle(paste(substr(hierarchy.name,1,10), substr(hierarchy.name,21,30), sep=" to ")) +
      ylab("log cluster coefficient") + xlab("log degree") +
      scale_x_continuous( labels=scaleFUN) + scale_y_continuous(labels=scaleFUN)

      selected.data <- hierarchy.df[hierarchy.df$selected.developer,]

      current.points.plot <- current.points.plot +
        geom_point(data = selected.data,shape = 15, size=4,
                   aes(jitter(hierarchy.log.deg, factor = 0.5),
                       colour = developer.names,
                       jitter(hierarchy.log.cc, factor = 0.5)))

      return(current.points.plot)
  })

  # save the plots
  pdf(file=plot.path, height= 6*ceiling(length(plots)/5), width = 10*5)
  do.call("grid.arrange", c(plots, ncol=5, newpage=TRUE))
  dev.off()
}

## mark core developers -------------------------------------------------------

#' This function plots the hierarchy data with one plot per time range and
#' marks the core developer, computed with the eigenvector centrality,
#' with a other color then the peripheral developers.
#' The core/peripheral classification is computed over the network of the
#' whole history and not for each time range separatly.
#'
#' @param hierarchy.data list of hierarchy data with one hierarchy data set per
#'                       time range
#' @param networks list of author networks
#' @param plot.path file path, where the result plot is saved
plot.with.marked.core.developers <- function(hierarchy.data, networks, number.developers, plot.path) {
  # compute the core/peripheral developer based on eigenvector centrality
  networks.with.eigen.centrality <- lapply(networks, function(net) {
    return(get.author.class.by.type(network = net,
                                    type = c("network.eigen")))
  })

  # merge network data with core/peripheral classification with hierarchy data
  hierarchy.with.core <- lapply(names(hierarchy.data), function(range.name){
    core.developers <- networks.with.eigen.centrality[[range.name]]$core
    hierarchy.df <- hierarchy.data[[range.name]]
    hierarchy.df$developer.names <- row.names(hierarchy.df)

    hierarchy.df$classification <- "peripheral"
    if (!is.null(core.developers) && !is.null(hierarchy.df$developer.names) &&
        nrow(core.developers) > 0) {
      hierarchy.df[hierarchy.df$developer.names %in% core.developers$author.name,
                   "classification"] <- "core"
    }
    return(hierarchy.df)
  })

  names(hierarchy.with.core)  <- names(hierarchy.data)
  # plot the hierarchy with marked
  plots <- lapply(names(hierarchy.with.core), function(range.name){
    hierarchy <- hierarchy.with.core[[range.name]]
    num.dev <- number.developers[range.name,]

    if (!is.null(hierarchy) && !is.null(nrow(hierarchy)) && row(hierarchy) > 0) {
      plot.with.core <- ggplot(data = hierarchy) +
        geom_point( aes(jitter(hierarchy.log.deg, factor = 0.5),
                        colour = classification,
                        jitter(hierarchy.log.cc, factor = 0.5))) +
        ggtitle(paste(substr(range.name,1,10), substr(range.name,21,30), sep=" to ")) +
        figureargs +  ylab("log cluster coefficient") + xlab("log degree") +
        scale_x_continuous( labels=scaleFUN) + scale_y_continuous(labels=scaleFUN) +
        xlab(paste("log degree \n  0 contacts: ", round(num.dev$num.deg.zero.percent, digits = 1),
                   " %, 1 contact: ", round(num.dev$num.deg.one.percent, digits = 1),
                   " %, >1 contact: ", round(num.dev$num.deg.greater.one.percent, digits = 1),
                   " %", sep="")) + theme(legend.title=element_blank(), legend.position="bottom")
      return(plot.with.core)
    }
  })

  plots <- plyr::compact(plots)

  # save plots
  pdf(file=plot.path, height= 7*ceiling(length(plots)/5), width = 10*5)
  do.call("grid.arrange", c(plots, ncol=5, newpage=TRUE))
  dev.off()
}


##  neighborhood of selected developer/ ego network ---------------------------

#' This function computes and plots the neighbourhood of one selected
#' developer. It marks the selected developer and his neighbours in
#' the known hierarchy plots.
#'
#' @param selected.developer name of the selected developer, whose neighbours
#'                           are visualized
#' @param hierarchy.data list of hierarchy data set for each time range
#' @param networks list of author networks, which are used to detect the
#'                 neighbours of an developer
#' @param plot.path file path, where the plots are saved
plot.neighborhood.of.selected.developer <- function(selected.developer,
                                                    hierarchy.data, networks,
                                                    plot.path){
  # compute egonetworks/neighbourhood of a selected developer
  hierarchy.with.ego.network <- lapply(names(networks), function(net.name){

    net <- networks[[net.name]]
    hierarchy <- hierarchy.data[[net.name]]
    if (!is.null(hierarchy) & !is.null(net)) {
      hierarchy$developer.names <- row.names(hierarchy)
      hierarchy$is.neighbour <- "all others"
      hierarchy$range <- net.name
      #hierarchy$is.selected.developer <- FALSE

      if (selected.developer %in% names(V(net))){
        selected.node <- V(net)[[selected.developer]]
        ego.network.list <- igraph::make_ego_graph(net, order=1,
                                                   nodes = selected.node,
                                                   mode = "all", mindist = 1)
        neighbour.nodes <- V(ego.network.list[[1]])

        logging::logdebug(paste("Number of vertices in network: ",
                                igraph::vcount(ego.network.list[[1]])))
        if (igraph::vcount(ego.network.list[[1]]) > 0 && !is.null(nrow(hierarchy))) {
          hierarchy[hierarchy$developer.names %in% neighbour.nodes$name,
                    "is.neighbour"] <- "neighbor"
        }
      }
      # for debugging
      #if (!is.null(hierarchy)) {
       # print(paste("Number of developers in ", net.name, " as neighbours: ",
      #              length(hierarchy[hierarchy$is.neighbour == "neighbor", "developer.names"]) ))
      #}
      return(hierarchy)
    }
  })

  # plot the hierarchy plots with marked selected developer and marked neighbours
  plots <- lapply(hierarchy.with.ego.network, function(hierarchy){
    if (!is.null(hierarchy) && !is.null(ncol(hierarchy))) {
      plot.with.ego <- ggplot(hierarchy) +
        geom_point(data = hierarchy, aes(jitter(hierarchy.log.deg, factor = 0.5),
                                         colour = is.neighbour, shape = is.neighbour,
                                         jitter(hierarchy.log.cc, factor = 0.5))) +
        scale_color_manual(breaks = c("all others", "neighbor"),
                           values = c("black", "red")) +
        scale_shape_manual(breaks = c("all others", "neighbor"),
                           values = c(16, 17))

      plot.with.ego <- plot.with.ego +
        geom_point(data = hierarchy[hierarchy$developer.names == selected.developer,],
                   shape= 8, size=5, colour="red",
                   aes(jitter(hierarchy.log.deg, factor = 0.5),
                       jitter(hierarchy.log.cc, factor = 0.5))) +
        figureargs + theme(legend.title=element_blank()) +
        ggtitle(paste(substr(hierarchy$range[[1]],1,10),
                      substr(hierarchy$range[[1]],21,30), sep=" to ")) +
        ylab("log cluster coefficient") + xlab("log degree") +
        theme(legend.position="bottom") +
        scale_x_continuous(labels=scaleFUN) + scale_y_continuous(labels=scaleFUN)

      return(plot.with.ego)
    }
  })

  plots <- plyr::compact(plots)
  # save plots
  pdf(file=plot.path, height= 7*ceiling(length(plots)/5), width = 10*5)
  do.call("grid.arrange", c(plots, ncol=5, newpage=TRUE))
  dev.off()

}

## aggregated analysis of ego networks ----------------------------------------

#' This function computes and plots the neighbourhood of one selected
#' developer. It marks the selected developer and his neighbours in
#' the known hierarchy plots.
#'
#' @param selected.developer name of the selected developer, whose neighbours
#'                           are visualized
#' @param hierarchy.data list of hierarchy data set for each time range
#' @param networks list of author networks, which are used to detect the
#'                 neighbours of an developer
#' @param plot.path file path, where the plots are saved
plot.aggregated.neighborhood.of.selected.developer <- function(selected.developer,
                                                               hierarchy.data, networks,
                                                               hierarchy.data.with.normalized.criteria,
                                                               number.of.steps = NUMBER.OF.STEPS,
                                                               plot.path){
  # compute egonetworks/neighbourhood of a selected developer
  decomposition.of.neighborhood <- lapply(names(networks), function(net.name){

    net <- networks[[net.name]]  # network
    hierarchy <- hierarchy.data[[net.name]] # hierarchy data
    if (!is.null(hierarchy) & !is.null(net) &
        net.name %in% names(hierarchy.data.with.normalized.criteria)) {
      hierarchy$developer.names <- row.names(hierarchy)
      hierarchy$is.neighbour <- "all others"
      hierarchy$range <- net.name
      #hierarchy$is.selected.developer <- FALSE

      neighbour.nodes <- NULL
      if (selected.developer %in% names(V(net))){
        # node of selected developer
        selected.node <- V(net)[[selected.developer]]

        # ego network
        ego.network.list <- igraph::make_ego_graph(net, order=1,
                                                   nodes = selected.node,
                                                   mode = "all", mindist = 1)
        neighbour.nodes <- V(ego.network.list[[1]])

        hierarchy[hierarchy$developer.names %in% neighbour.nodes$name,
                  "is.neighbour"] <- "neighbor"
      }

      hierarchy <- compute.division.into.hierarchical.and.non.hierachical.part(hierarchy, NUMBER.OF.STEPS,
                                              hierarchy.data.with.normalized.criteria,
                                              net.name)

      # select number of neighbors in the hierarchically structured part and remaining part
      num.neighbours.in.hierarchy <- nrow(hierarchy[hierarchy$part.of.hierarchy &
                                                    hierarchy$is.neighbour == "neighbor" ,])
      num.neighbours.not.in.hierarchy <- nrow(hierarchy[(!hierarchy$part.of.hierarchy) &
                                                         hierarchy$is.neighbour == "neighbor" ,])

      return(data.frame("neighbors.in.hierarchy"=num.neighbours.in.hierarchy,
                        "neighbors.not.in.hierachy"=num.neighbours.not.in.hierarchy,
                        "neighbors"=length(neighbour.nodes),
                        "range"=net.name))
    }
  })

  neighborhood <- rbind.fill(decomposition.of.neighborhood)
  neighborhood$range.number <- c(1:nrow(neighborhood))

  subdata <- neighborhood[,c('neighbors.in.hierarchy', 'neighbors.not.in.hierachy','range.number')]
  df <- melt(subdata, id.vars = 'range.number')

  # Data for one time step cannot be plotted with "geom_area", so make two time steps out of it to be plotted
  available.data.for.ranges = unique(df$range.number[which(df$value != 0)])
  if (length(available.data.for.ranges) < 2 & length(available.data.for.ranges) > 0) {
    if (df[available.data.for.ranges, "range.number"] < df[available.data.for.ranges + 1, "range.number"]) {
      df[available.data.for.ranges+1, "value"] = df[available.data.for.ranges, "value"]
      df[(available.data.for.ranges+1) + max(df$range.number), "value"] = df[available.data.for.ranges + max(df$range.number), "value"]
    }
  }

  plot.aggregated.neighborhood <- ggplot(data=df,
                                         aes(x=range.number, y=value, group=variable, fill=variable)) +
    geom_area(position="fill") +
    figureargs + theme(legend.position="bottom", axis.title.y=element_blank()) +
    theme(legend.title=element_blank()) +
    scale_fill_discrete(name  ="", breaks=c("neighbors.in.hierarchy", "neighbors.not.in.hierachy"),
                        labels=c("Perc. in hierachy  ", "Perc. not in hierarchy  ")) +
    scale_y_continuous(labels= function(x) paste0(100*x,'%')) +
    scale_fill_manual(values = c( "lightblue", "darkblue"), name  ="",
                      breaks=c("neighbors.in.hierarchy", "neighbors.not.in.hierachy"),
                      labels=c("Perc. in hierachy  ", "Perc. not in hierarchy  ")) +
    xlab("time steps")


  print(plot.aggregated.neighborhood)
  # save plot
  pdf(file=plot.path, height= 6, width = 10)
  print(plot.aggregated.neighborhood)
  dev.off()

}


## mail count --------------------
#TODO not finished

#' The function plots the hierarchy data with point size regarding the number of
#' e-mail a developer writes. The list of plots is saved in the file with the
#' file path \code{plot.path}.
#'
#' @param network.list list of networks per time range
#' @param project.data project data object (necessary to compute the number of
#'                     edited artifacts)
#' @param activity selected artifact type (not yet used)
#' @param hierarchy.data list of hierarchy data with a data set for each time
#'                       range
#' @param plot.path file path for the result file with the hierarchy plots
hierarchy.plot.with.artifact.count.mail <- function(network.list, project.data, activity,
                                                    hierarchy.data, plot.path) {
  # count artifacts per developer
  networks <- add.vertex.attribute.author.email(network.list,
                                                project.data,
                                                name = "author.email",
                                                default.value = NA)

  # merge hierarchy and artifact count data
  #TODO: e-mails
  hierarchies.with.artifact.count <- lapply(names(hierarchy.data), function(project.name) {
    current.hierarchy.data <- hierarchy.data[[project.name]]
    current.network <- networks[[project.name]]
    current.network.df.vertices <- as_data_frame(current.network, "vertices")

    hierarchy.with.artifact.count <- lapply(row.names(current.hierarchy.data),
                                            function(developer.name){
      if (!is.null(current.network.df.vertices) && nrow(current.network.df.vertices) > 0) {
        current.hierarchy.data[developer.name, "artifact.count"] <-
          current.network.df.vertices[developer.name,"artifact.count"]
        return(current.hierarchy.data[developer.name,])
      } else {
        logging::loginfo(paste("hierarchy.plot.with.artifact.count: ",
                               "No artifact.count for developer ", developer.name, "."))
        return(NULL)
      }
    })
    return(hierarchy.with.artifact.count)
  })

  names(hierarchies.with.artifact.count) <- names(hierarchy.data)

  hierarchies.with.artifact.count <- plyr::compact(hierarchies.with.artifact.count)

  # plot hierarchy with artifact count data
  plots <- lapply(names(hierarchies.with.artifact.count), function(hierarchy.name){
    hierarchy <- hierarchies.with.artifact.count[[hierarchy.name]]
    if (!is.null(hierarchy) && length(hierarchy) > 0) {
      hierarchy.df <- rbind.fill(hierarchy)
      current.points.plot <- ggplot() +
        geom_point(data = hierarchy.df, aes(jitter(hierarchy.log.deg, factor = 0.5),
                                            size = artifact.count, colour=artifact.count,
                                            jitter(hierarchy.log.cc, factor = 0.5))) +
        figureargs +
        ggtitle(paste(substr(hierarchy.name,1,10), substr(hierarchy.name,21,30), sep=" to ")) +
        ylab("log cluster coefficient") + xlab("log degree") +
        scale_x_continuous( labels=scaleFUN) + scale_y_continuous(labels=scaleFUN) +
        theme(legend.position="right", legend.box = "vertical") +
        scale_size_continuous(name="number edited files")
      return(current.points.plot)
    }
  })

  plots <- plyr::compact(plots)

  # save plots
  pdf(file=plot.path, height= 7*ceiling(length(plots)/5), width = 12*5)
  do.call("grid.arrange", c(plots, ncol=5, newpage=TRUE))
  dev.off()
}
