## (c) Barbara Eckl-Ganser, July 2018 - October 2020
## ecklbarb@fim.uni-passau.de
## Methods to analyse hierarchy

## LOAD NEEDED LIBRARIES

library("igraph")
library("ggplot2")
#library("gplots")
library("gridExtra")
library("reshape2")

# break points
#library("freeknotsplines")

scaleFUN <- function(x) sprintf("%.1f", x)

figureargs <- theme_light(base_size = 25)

## THRESHOLDS FOR DETECTION OF HIERARCHICAL PART

# maximum residual variance of the regression line
MAX.VARIANCE = 0.5

# minimum angle between x-axis and regression line
MIN.ANGLE = 35

# minimum percentage of developers in the hierarchical part
MIN.PERC.DEVS.IN.HIERARCHICAL.PART = 0.05

# number of buckets (intervals) of equal length
NUMBER.OF.STEPS = 20


## Stepwise computing group of nodes, angle, variance -------------------------------

#' The method computes for a increasing number of nodes the linear regression, the angle of the line and the
#' x-axis and the variance of the including points.
#'
#' @param hierarchy.df data frame, which includes the information about node degree and cluster coefficient;
#'                        computed with the `metrics.hierarchy`
#' @param number.of.steps number of steps in the iterative process of finding the hierarchical structure;
#'                       the default value is 20 steps
#'
#' @return data frame including angle, variance, number of nodes in the hierarchical structure, percentage of
#'         nodes in the hierarchical structure and the gradient of the line of the linear regression
compute.hierarchical.parts <- function(hierarchy.df, number.of.steps = NUMBER.OF.STEPS) {
  logging::loginfo("Starting computing stepwise hierarchy parameters.")
  if (is.null(hierarchy.df) || is.na(hierarchy.df) || is.null(nrow(hierarchy.df))) {
    logging::loginfo("No hierarchy data for this range.")
    return(NULL)
  }
  x.minimum = min(hierarchy.df$hierarchy.log.deg) # Minimum node degree in the data frame
  x.maximum = max(hierarchy.df$hierarchy.log.deg) # Maximum node degree in the data frame
  y.minimum = min(hierarchy.df$hierarchy.log.cc)  # Minimum cluster coefficient in the data frame
  y.maximum = max(hierarchy.df$hierarchy.log.cc)  # Maximum cluster coefficient in the data frame

  # Step size
  x.step.width = (x.maximum - x.minimum)/ number.of.steps
  #y.step.width = (y.maximum - y.minimum)/ number.of.steps

  hierarchy.analysis = data.frame()
  for (current.step in 1:number.of.steps) {
    # selection of points which are currently used for the hierarchicaly part
    # of the points in the hierarchy plot
    #current.points <- hierarchy.df[hierarchy.df$hierarchy.log.cc <= current.step * y.step.width + y.minimum, ]
    current.points <- hierarchy.df[hierarchy.df$hierarchy.log.deg > x.maximum - current.step * x.step.width, ]

    if (!is.null(current.points) && nrow(current.points) > 1) {
      reg <- lm(hierarchy.log.cc~hierarchy.log.deg, data = current.points) # compute linear regression
      reg.gradient <- c(1, reg$coefficients [2])
      sigma.square <- mean(resid(reg)^2)  # variance

      angle <- rad2deg(atan(reg.gradient))  # angle with teh xaxis

      hierarchy.analysis <- rbind(hierarchy.analysis, c("current.step" = current.step,
                                                        "number.all.developer" = nrow(hierarchy.df),
                                                        "n" = nrow(current.points),
                                                        "percent.n" = (nrow(current.points)/nrow(hierarchy.df)),
                                                        "variance" = sigma.square,
                                                        "gradient" = reg.gradient["log.deg"],
                                                        "angle" = angle[2]
                                                      ))

      # current.points.plot <- ggplot(current.points) + geom_point(size = 0.3, aes(jitter(deg, factor = 0.5), jitter(cc, factor = 0.5))) +  theme(legend.position="none") +
      #   geom_smooth(aes(jitter(deg, factor = 0.5), jitter(cc, factor = 0.5)), method='lm',formula=y~x, size = 0.4) + #+ geom_histogram(aes(log.deg))
      #   scale_y_log10() + scale_x_log10()
      # print(current.points.plot)
    }
  }

  if(length(hierarchy.analysis) != 0){
    names(hierarchy.analysis) <- c("cur.step","number.all.developer", "n", "percent.n", "variance", "gradient", "angle")

    hierarchy.analysis$perc.n.diff <- NA
    hierarchy.analysis$vari.diff <- NA
    hierarchy.analysis$angle.diff <- NA
    if (nrow(hierarchy.analysis) > 1) {
      for (i in 2:nrow(hierarchy.analysis)) {
        hierarchy.analysis[i,]$perc.n.diff <- hierarchy.analysis[i,]$percent.n - hierarchy.analysis[i-1,]$percent.n
        hierarchy.analysis[i,]$vari.diff <- hierarchy.analysis[i,]$variance - hierarchy.analysis[i-1,]$variance
        hierarchy.analysis[i,]$angle.diff <- hierarchy.analysis[i,]$angle - hierarchy.analysis[i-1,]$angle
      }
    }
    logging::loginfo("Finished computing hierarchy parameter")
    return(hierarchy.analysis)
  }

}

#' similar to 'compute.hierarchical.parts', but plots the the linear regression functions
#' specific case of \code{plot.hierarchy.only.realistic.values}
plot.hierarchy.with.different.regressions <- function(hierarchy.df, number.of.steps = NUMBER.OF.STEPS) {
  x.minimum = min(hierarchy.df$hierarchy.log.deg) # Minimum node degree in the data frame
  x.maximum = max(hierarchy.df$hierarchy.log.deg) # Maximum node degree in the data frame
  y.minimum = min(hierarchy.df$hierarchy.log.cc)  # Minimum cluster coefficient in the data frame
  y.maximum = max(hierarchy.df$hierarchy.log.cc)  # Maximum cluster coefficient in the data frame

  # Step size
  x.step.width = (x.maximum - x.minimum)/ (number.of.steps)
  y.step.width = (y.maximum - y.minimum)/ number.of.steps

  current.points.plot <- ggplot() +
    geom_point(data = hierarchy.df, size = 0.3, aes(jitter(hierarchy.log.deg, factor = 0.5), jitter(hierarchy.log.cc, factor = 0.5))) +
    theme(legend.position="none") + xlab("log.degree") + ylab("log.clus.coef")
    #scale_y_log10() + scale_x_log10()

  color.values=colorRampPalette(c("yellow","red", "blue"))(number.of.steps)
  color.rectangle=colorRampPalette(c("green","blue"))(number.of.steps)

  color.step = 1
  for (current.step in 1:(number.of.steps)) {
    # selection of points which are currently used for the hierarchicaly part
    # of the points in the hierarchy plot
    #current.points <- hierarchy.df[hierarchy.df$hierarchy.log.cc <= current.step * y.step.width + y.minimum, ]
    current.x.limit <- x.maximum - current.step * x.step.width
    current.points <- hierarchy.df[hierarchy.df$hierarchy.log.deg >=  current.x.limit, ]

    if (!is.null(current.points) && nrow(current.points) > 1) {
      reg <- lm(hierarchy.log.cc~hierarchy.log.deg, data = current.points) # compute linear regression
      reg.gradient <- c(1, reg$coefficients[2])
      sigma.square <- mean(resid(reg)^2)  # variance

      angle <- rad2deg(atan(reg.gradient))  # angle with teh xaxis

      current.points.plot <- current.points.plot +
        geom_smooth(data = current.points, aes(hierarchy.log.deg, hierarchy.log.cc),
                    method='lm',formula=y~x, size = 0.4, se = FALSE, colour = color.values[color.step]) +
        #geom_hline(yintercept = current.step * y.step.width + y.minimum, colour = color.rectangle[color.step]) #+
         geom_vline(xintercept = current.x.limit, colour = color.rectangle[color.step]) +
         geom_text(aes(y=0), x=current.x.limit, label=current.step,
                   colour = color.values[color.step],
                   angle = 90, vjust = 1.2, size=3)

      color.step <- color.step + 1
    }

  }
  return(current.points.plot)
}

#' The method computes and plots for a increasing number of nodes the linear regression,
#' the angle of the line and the
#' x-axis and the variance of the including points (on the right side of the plot).
#' NOTE: the method is similar to `compute.hierarchical.parts`
#'
#' @param hierarchy.df data frame, which includes the information about node degree and cluster coefficient;
#'                        computed with the `metrics.hierarchy`
#' @param number.of.steps number of steps in the iterative process of finding the hierarchical structure;
#'                       the default value is 20 steps
#' @param all.lines parameter, whether all lines are plotted or only the lines of the realistic values [default: TRUE]
#' @param all.lines.colored if all lines are plotted, there [default: FALSE]
#' @param with.left.group describe, whether the left group should be plotted, too [default: FALSE]
#' @param vertical.lines plots only the hierarchy dots and the vertical lines, can be combined with all.lines.colored
#'                       [default: FALSE]
#'
#' @return data frame including angle, variance, number of nodes in the hierarchical structure, percentage of
#'         nodes in the hierarchical structure and the gradient of the line of the linear regression
plot.hierarchy.only.realistic.values <- function(hierarchy.df,
                                                 number.of.steps = NUMBER.OF.STEPS,
                                                 all.lines = TRUE,
                                                 all.lines.colored = FALSE,
                                                 with.left.group = FALSE,
                                                 only.selected.group = FALSE,
                                                 vertical.lines = FALSE) {
  logging::loginfo("Start, plotting of hierarchy with realistic regression line.")
  x.minimum = min(hierarchy.df$hierarchy.log.deg) # Minimum node degree in the data frame
  x.maximum = max(hierarchy.df$hierarchy.log.deg) # Maximum node degree in the data frame
  y.minimum = min(hierarchy.df$hierarchy.log.cc)  # Minimum cluster coefficient in the data frame
  y.maximum = max(hierarchy.df$hierarchy.log.cc)  # Maximum cluster coefficient in the data frame

  # Step size
  x.step.width = (x.maximum - x.minimum)/ (number.of.steps)
  y.step.width = (y.maximum - y.minimum)/ number.of.steps

  current.points.plot <- ggplot() +
    geom_point(data = hierarchy.df, size = 0.3, aes(jitter(hierarchy.log.deg, factor = 0.5), jitter(hierarchy.log.cc, factor = 0.5))) +
    theme(legend.position="none") + xlab("log.degree") + ylab("log.clus.coef")
  #scale_y_log10() + scale_x_log10()

  color.values=colorRampPalette(c("yellow","red", "blue"))(number.of.steps)
  color.rectangle=colorRampPalette(c("green","blue"))(number.of.steps)

  color.step = 1
  for (current.step in 1:(number.of.steps)) {
    # selection of points which are currently used for the hierarchical part
    # of the points in the hierarchy plot
    #current.points <- hierarchy.df[hierarchy.df$hierarchy.log.cc <= current.step * y.step.width + y.minimum, ]
    current.x.limit <- x.maximum - current.step * x.step.width
    current.points <- hierarchy.df[hierarchy.df$hierarchy.log.deg >=  current.x.limit, ]

    if (!is.null(current.points) && nrow(current.points) > 1) {
      reg <- lm(hierarchy.log.cc~hierarchy.log.deg, data = current.points) # compute linear regression
      reg.gradient <- c(1, reg$coefficients[2])
      sigma.square <- mean(resid(reg)^2)  # variance

      angle <- rad2deg(atan(reg.gradient))  # angle with the xaxis

      if (vertical.lines) {
        if (all.lines.colored) {
          current.points.plot <- current.points.plot +
            geom_vline(xintercept = current.x.limit, colour = color.rectangle[color.step]) +
            geom_text(aes(y = 0), x = current.x.limit, label = current.step,
                      colour = color.rectangle[color.step],
                      angle = 90, vjust = 1.2, size=3)
        } else {
          current.points.plot <- current.points.plot +
            geom_vline(xintercept = current.x.limit, colour = "grey") +
            geom_text(aes(y=0), x=current.x.limit, label=current.step,
                      colour = "grey",
                      angle = 90, vjust = 1.2, size=3)
        }

      } else {

        if (all.lines) {
          current.points.plot <- current.points.plot +
            #geom_hline(yintercept = current.step * y.step.width + y.minimum, colour = color.rectangle[color.step]) #+
            geom_vline(xintercept = current.x.limit, colour = color.rectangle[color.step])
        }

        is.realistic.value <- sigma.square < MAX.VARIANCE &&
                              (nrow(current.points)/nrow(hierarchy.df) > MIN.PERC.DEVS.IN.HIERARCHICAL.PART) &&
                              !is.na(angle[2]) &&
                              angle[2] < -MIN.ANGLE

        if (is.realistic.value) {
          current.points.plot <- current.points.plot +
            geom_smooth(data = current.points, aes(hierarchy.log.deg, hierarchy.log.cc),
                        method='lm',formula=y~x, size = 0.4, se = FALSE, colour = color.values[color.step]) +
            geom_text(aes(y=0), x=current.x.limit, label=current.step,
                      colour = color.values[color.step],
                      angle = 90, vjust = 1.2, size=3)
        } else {
          if (all.lines) {
            if (all.lines.colored) {
              current.points.plot <- current.points.plot +
                geom_smooth(data = current.points, aes(hierarchy.log.deg, hierarchy.log.cc),
                            method='lm',formula=y~x, size = 0.4, se = FALSE, colour = color.values[color.step]) +
                geom_text(aes(y = 0), x = current.x.limit, label = current.step,
                          colour = color.values[color.step],
                          angle = 90, vjust = 1.2, size=3)
            } else {
              current.points.plot <- current.points.plot +
                geom_smooth(data = current.points, aes(hierarchy.log.deg, hierarchy.log.cc),
                            method='lm',formula=y~x, size = 0.4, se = FALSE, colour = "grey") +
                geom_text(aes(y=0), x=current.x.limit, label=current.step,
                          colour = "grey",
                          angle = 90, vjust = 1.2, size=3)
            }
          }
        }

        if (with.left.group) {
          left.group.points <- hierarchy.df[hierarchy.df$hierarchy.log.deg <  current.x.limit, ]
          if (all.lines) {
            if (all.lines.colored) {
              current.points.plot <- current.points.plot +
                geom_smooth(data = left.group.points, aes(hierarchy.log.deg, hierarchy.log.cc),
                            method='lm',formula=y~x, size = 0.4, se = FALSE, colour = color.values[color.step])
            } else {
              if (is.realistic.value) {
                current.points.plot <- current.points.plot +
                  geom_smooth(data = left.group.points, aes(hierarchy.log.deg, hierarchy.log.cc),
                              method='lm',formula=y~x, size = 0.4, se = FALSE, colour = color.values[color.step])
              } else {
                current.points.plot <- current.points.plot +
                  geom_smooth(data = left.group.points, aes(hierarchy.log.deg, hierarchy.log.cc),
                              method='lm',formula=y~x, size = 0.4, se = FALSE, colour = "grey")
              }
            }
          } else {
            if (is.realistic.value) {
              current.points.plot <- current.points.plot +
                geom_smooth(data = left.group.points, aes(hierarchy.log.deg, hierarchy.log.cc),
                            method='lm',formula=y~x, size = 0.4, se = FALSE, colour = color.values[color.step])
            }
          }
        }
      }
      color.step <- color.step + 1
    }

  }
  logging::loginfo("Finished, plotting of hierarchy with realistic regression line.")
  return(current.points.plot)
}

#' The method plots for each time window, network, the angle and variance of the iterative growing
#' number of points in the hierarchical structure.
#'
#'@param angle.sigma.data data frame, which includes the information about angle and sigma;
#'                        computed with the `compute.hierarchical.parts`
#'@param plot.output.file path and name of the output file with a plot per time window (network) and
#'                        different ratio of points in the hierarchical part
plot.angle.variance.sigma <- function(angle.variance.sigma.data, plot.output.file) {
  logging::loginfo("Started, plotting of hierarchy parameter in stepwise process.")
  plots <- lapply(angle.variance.sigma.data, function(data.one.time) {
    maximum.variance = max(data.one.time$variance)
    if(!is.null(data.one.time) && nrow(data.one.time) !=0) {
      xdf <- cbind("time"=1:nrow(data.one.time),data.one.time[,c("percent.n", "variance", "angle")])
      current.plot <- ggplot(data = xdf, aes(x=time)) +  xlab("time") +
        geom_line(aes(y = angle, colour = "Angle(blue)")) +
        geom_line(aes(y = variance/(maximum.variance*0.01), colour = "sigma")) +
        scale_y_continuous(sec.axis = sec_axis(~.*0.01, name = "Var(red), %nodes(green)")) +
        geom_line(aes(y = percent.n/0.01, colour = "vertices")) +
        scale_colour_manual(values = c("blue", "red", "green")) +
        labs(y = "Angle(blue)[°]",x = "iteration step")  +
        theme(legend.position="none", axis.text=element_text(size=6),
              axis.title=element_text(size=8))

      return(current.plot)
    }
  })

  logging::loginfo("Finishing, plotting of hierarchy parameter in stepwise process.")
  plots <- plots[2: (length(plots)-1)]
  pdf(file=plot.output.file, height= 1.5*(ceiling(length(plots)/4)))
  do.call("grid.arrange", c(plots, ncol=4, newpage=TRUE))
  dev.off()

}

#' Method plots the meta data from the analysis with
#' variance, angle, percentage of nodes
#'
#' @param network.data meta data about hierarchy
#'
#' @return one plot with variance, angle and percentage of nodes for the different steps
only.one.plot.angle.variance.sigma <- function(network.data) {
  logging::loginfo("Started, plotting of hierarchy parameter in stepwise process.")
  maximum.variance = max(network.data$variance)
  if(!is.null(network.data) && nrow(network.data) !=0) {
    xdf <- cbind("time"=1:nrow(network.data), network.data[,c("percent.n", "variance", "angle")])
    current.plot <- ggplot(data = xdf, aes(x=time)) +  xlab("time") +
      geom_line(aes(y = angle, colour = "Angle(blue)")) +
      geom_line(aes(y = variance/(maximum.variance*0.01), colour = "sigma")) +
      scale_y_continuous(sec.axis = sec_axis(~.*0.01, name = "Var(red), %nodes(green)")) +
      geom_line(aes(y = percent.n/0.01, colour = "vertices")) +
      scale_colour_manual(values = c("blue", "red", "green")) +
      labs(y = "Angle(blue)[°]",x = "iteration step")  +
      theme(legend.position="none")#, axis.text=element_text(size=6),
            #axis.title=element_text(size=8))

    return(current.plot)
  }
  logging::loginfo("Finished, plotting of hierarchy parameter in stepwise process.")
}

#' Computes the degree [°] of an angle given a rational number.
#'
#' @param rad angle as a rational number
#'
#' @return angle in degree
rad2deg <- function(rad) {(rad * 180) / (pi)}

## COMPUTE DIFFERENT K-MEANS CLUSTER ALGORITHMS AND BREAKPOINT DETECTION METHODS -------------

#' Compute the k-means with two centers to divide the developers in two groups using three different algorithms
#' and compute a breakpoint to do the same basing on the hierarchy measure which include
#' node degree and cluster coefficient.
#'
#' @param hierarchy data frame which include the cluster coefficient and the node degree
#'                  of each developer of a author network (hierarchy)
#'
#' @return hierarchy with k-means clustering and breakpoint detection information and additional informations,
#' whether the developer was already a developer in the previous network
compute.kmeans.and.breakpoint.detection <- function(hierarchy) {
  logging::loginfo("Started, computing kmeans and breakpoint detection on hierarchy data.")
  hierarchy.with.kmeans.clustering <- lapply(1:length(hierarchy), function(network_num){

    ## previous in the network
    hierarchy_df <- hierarchy[[network_num]]
    if (!is.null(hierarchy_df) && nrow(hierarchy_df) > 0) {
      hierarchy_df$already.in.previos <- FALSE
      if (network_num > 1) {
        hierarchy_df[which(is.element(rownames(hierarchy_df),
                                      intersect(rownames(hierarchy_df), rownames(hierarchy[[network_num -1]])))),
                     "already.in.previos"] <- TRUE
      }

      # k-mean clustering algorithms (Hartigan-Wong, Lloyd, MacQueen)
      clu <- kmeans(data.frame(hierarchy_df$hierarchy.log.deg,
                               hierarchy_df$hierarchy.log.cc), centers=2, #data.frame("log.deg"=c(1.0, 3.0), "log.cc"=c(0.0, -1.0)),
                    iter.max = 15)
      clu_Lloyd <- kmeans(data.frame(hierarchy_df$hierarchy.log.deg,
                                     hierarchy_df$hierarchy.log.cc), centers=2, algorithm = "Lloyd",
                          iter.max = 15)
      clu_MacQueen <- kmeans(data.frame(hierarchy_df$hierarchy.log.deg,
                                        hierarchy_df$hierarchy.log.cc), centers=2, algorithm = "MacQueen",
                             iter.max = 15)
      hierarchy_df$clustering <- clu$cluster
      hierarchy_df$clustering_Lloyd <- clu_Lloyd$cluster
      hierarchy_df$clustering_MacQueen <- clu_MacQueen$cluster

      ## breakpoint detection
      xy.freekt <- freelsgen(hierarchy_df$hierarchy.log.deg, hierarchy_df$hierarchy.log.cc, degree = 1, numknot = 1) # numknot = 3?? (inklusive Start- und Endpunkt)
      optknot <- xy.freekt@optknot

      left.right.group <- cut(hierarchy_df$hierarchy.log.deg,
                              breaks = c(-Inf, optknot, Inf),
                              labels = c("left", "right"))

      hierarchy_df$left.right.group <- "left"
      hierarchy_df[which(hierarchy_df$hierarchy.log.deg > optknot), "left.right.group"] <- "right"
      return(hierarchy_df)
    }
  })

  logging::loginfo("Finishing, computing kmeans and breakpoint detection on hierarchy data.")
  return(hierarchy.with.kmeans.clustering)
}


#' The method separates the points of the hierarchy in two groups different k-means algorithms
#' and breakpoint detection methods are used. Additionally, the variance in the groups is
#' computed.
#'
#' @param hierarchy.with.clustering data frame which include the cluster coefficient and the node degree
#'                  of each developer of a author network (hierarchy) and additional the information
#'                  about grouping based on different methods (k-means algorithms and breakpoint detection)
#' @return hierarchy data with additional information about the angles and variance of
#'         different clusterings
compute.angles.and.variance.of.different.groupings <- function(hierarchy.with.clustering) {
  logging::loginfo("Started, computing angles between kmeans and breakpoint detected groups on hierarchy data.")

  ## iterates over the hierarchy data of the (time-based) networks
  hierarchy.with.clustering.and.angles <- lapply(hierarchy.with.clustering, function(nw){
    ## freespline breakpoint
    if (!is.null(nw)) {
      nw$sigma.square.right.bk <- NaN
      nw$sigma.square.left.bk <- NaN
      nw$angle.bk <- NaN

      if(!is.null(nw[nw$left.right.group == "right", ])) {
        reg <- lm(hierarchy.log.cc~hierarchy.log.deg, data = nw[nw$left.right.group == "right",])
        reg.gradient.right <- c(1, reg$coefficients [2])
        nw$sigma.square.right.bk <- mean(resid(reg)^2) # variance
      }
      if(!is.null(nw[nw$left.right.group == "left", ])) {
        reg <- lm(hierarchy.log.cc~hierarchy.log.deg, data = nw[nw$left.right.group == "left",])
        reg.gradient.left <- c(1, reg$coefficients [2])
        nw$sigma.square.left.bk <- mean(resid(reg)^2)
      }

      nw$angle.bk <- acos(sum(reg.gradient.left*reg.gradient.right) / (sqrt(sum(reg.gradient.left * reg.gradient.left))
                                                                       * sqrt(sum(reg.gradient.right * reg.gradient.right))))
      nw$angle.bk <- rad2deg(nw$angle.bk)

      ## k-means clustering
      nw$sigma.square.right.cl <- NaN
      nw$sigma.square.left.cl <- NaN
      nw$angle.cl <- NaN

      if(!is.null(nw[nw$clustering == 1, ])) {
        reg <- lm(hierarchy.log.cc~hierarchy.log.deg, data = nw[nw$clustering == 1, ])
        reg.gradient.right <- c(1, reg$coefficients [2])
        nw$sigma.square.right.cl <- mean(resid(reg)^2) # variance
      }
      if(!is.null(nw[nw$clustering == 2, ])) {
        reg <- lm(hierarchy.log.cc~hierarchy.log.deg, data = nw[nw$clustering == 2, ])
        reg.gradient.left <- c(1, reg$coefficients [2])
        nw$sigma.square.left.cl <- mean(resid(reg)^2)
      }

      theta.cl <- acos(sum(reg.gradient.left*reg.gradient.right) / (sqrt(sum(reg.gradient.left * reg.gradient.left))
                                                                    * sqrt(sum(reg.gradient.right * reg.gradient.right))))

      nw$angle.cl <- rad2deg(theta.cl)

      # k-means clustering Lloyd
      nw$sigma.square.right.lloyd <- NaN
      nw$sigma.square.left.lloyd <- NaN
      nw$angle.lloyd <- NaN

      if(!is.null(nw[nw$clustering_Lloyd == 1, ])) {
        reg <- lm(hierarchy.log.cc~hierarchy.log.deg, data = nw[nw$clustering_Lloyd == 1, ])
        reg.gradient.right <- c(1, reg$coefficients [2])
        nw$sigma.square.right.lloyd <- mean(resid(reg)^2) # variance
      }
      if(!is.null(nw[nw$clustering_Lloyd == 2, ])) {
        reg <- lm(hierarchy.log.cc~hierarchy.log.deg, data = nw[nw$clustering_Lloyd == 2, ])
        reg.gradient.left <- c(1, reg$coefficients [2])
        nw$sigma.square.left.lloyd <- mean(resid(reg)^2)
      }

      theta.lloyd <- acos( sum(reg.gradient.left*reg.gradient.right) / ( sqrt(sum(reg.gradient.left * reg.gradient.left))
                                                                         * sqrt(sum(reg.gradient.right * reg.gradient.right))))
      nw$angle.lloyd <- rad2deg(theta.lloyd)

      # k-means clustering MacQueen
      nw$sigma.square.right.mq <- NaN
      nw$sigma.square.left.mq <- NaN
      nw$angle.mq <- NaN

      if(!is.null(nw[nw$clustering_MacQueen == 1, ])) {
        reg <- lm(hierarchy.log.cc~hierarchy.log.deg, data = nw[nw$clustering_MacQueen == 1, ])
        reg.gradient.right <- c(1, reg$coefficients [2])
        nw$sigma.square.right.mq <- mean(resid(reg)^2) # variance
      }
      if(!is.null(nw[nw$clustering_MacQueen == 2, ])) {
        reg <- lm(hierarchy.log.cc~hierarchy.log.deg, data = nw[nw$clustering_MacQueen == 2, ])
        reg.gradient.left <- c(1, reg$coefficients [2])
        nw$sigma.square.left.mq <- mean(resid(reg)^2)
      }

      theta.mq <- acos(sum(reg.gradient.left*reg.gradient.right) / (sqrt(sum(reg.gradient.left * reg.gradient.left))
                                                                    * sqrt(sum(reg.gradient.right * reg.gradient.right))))
      nw$angle.mq <- rad2deg(theta.mq)

      return(nw)
    }
  })

  logging::loginfo("Finished, computing angles between kmeans and breakpoint detected groups on hierarchy data.")

  return(hierarchy.with.clustering.and.angles)
}

#' The function combines different plots to an overview of the evolution of hierarchy.
#'
#' @param hierarchy.data.set data set with hierarchy data
#' @param plot.output.file file path, where the output plot is saved
#'
#' @return one file with an grid layout with one row for each time window of the project.
plot.overview.hierarchy <- function(hierarchy.data.set, plot.output.file) {
  logging::loginfo("Started combining different plots to an overview.")
  plots <-list()
  for (hierarchy.data.num in 1:length(hierarchy.data.set)) {
    hierarchy.data <- hierarchy.data.set[[hierarchy.data.num]]
    plot.stepwise.hierarchy <- plot.hierarchy.only.realistic.values(hierarchy.data, all.lines = FALSE) +
      geom_label(aes(y=0.10), x=1, colour="black", label=hierarchy.data.num, size=3)
    plots[[as.character(5*hierarchy.data.num)]] <- plot.stepwise.hierarchy

    plot.all.lines <- plot.hierarchy.with.different.regressions(hierarchy.data)
    plots[[as.character(5*hierarchy.data.num + 1)]] <- plot.all.lines

    plot.breakpoints <- plot.stepwise.angle.related.breakpoint.detection(hierarchy.data)
    plots[[as.character(5*hierarchy.data.num + 2)]] <- plot.breakpoints

    hierarchy.analysis <- compute.hierarchical.parts(hierarchy.data )
    plot.angle.timeseries <- only.one.plot.angle.variance.sigma(hierarchy.analysis)
    plots[[as.character(5*hierarchy.data.num + 3)]] <- plot.angle.timeseries

    hierarchy.analysis <- data.frame(lapply(hierarchy.analysis, function(x){round(x,digits = 2)}))
    plots[[as.character(5*hierarchy.data.num + 4)]] <-  grid.arrange(tableGrob(hierarchy.analysis))
  }

  #plots <- unlist(plots)
  logging::loginfo("Finished combining different plots to an overview.")
  pdf(file=plot.output.file, height= 6*(ceiling(length(plots)/5)), width = 28)
  do.call("grid.arrange", c(plots, ncol=5, newpage=TRUE))
  dev.off()
}


## PLOT angle --------------------------------------------

#' Extracts and plots the angle between the linear regression of the hierarchical and the
#' non-hierarchical part of the hierarchy structure over time.
#'
#' Save the plot of the time series of the angles.
compute.and.plot.angle.between.hierarchical.and.nonhierarchical <- function(hierarchy.with.angles) {
  ## extract the data of angles between linear regressions of the hierarchical part and the
  ## non-hierarchical part of the hierarchy structure
  angles <- lapply(hierarchy_only_networks_with_edges, function(nw){
    return(data.frame("bk" = round(nw$angle.bk[[1]]), "Hartigan" = round(nw$angle.cl[[1]]),
                      "Lloyd" = round(nw$angle.lloyd[[1]]), "MacQueen" = round(nw$angle.mq[[1]])))
  })

  ## generate plot
  xdf <- reshape2::melt(angles)
  q <- ggplot(data = xdf, aes(L1, value, color = variable, shape = variable)) +
    geom_point() + geom_line()

  ## save plot
  ggsave(file=paste(OUTPUT.PATH, "breakpoints", paste(CASESTUDY, ANALYSIS, sep="-", "change_angles.pdf"), sep="/"),
         q, height=5)
}

## More breakpoints regarding the angle ----------------------------------------

#' This method plots the different linear regression of hierarchy data. If the difference
#' of angle between the right and the left regression line is more than 30°. The line is
#' printed and the algorithm starts with a new group. With this approach, we can split the
#' data, if the gradients are different enough.
#'
#' @param hierarchy.df hierarchy data
#' @param number.of.steps number of calculation steps (to increase the number of data
#' points from right to the left) [default: 20]
#'
#' @return plot with angle at the breakpoint as additional information
plot.stepwise.angle.related.breakpoint.detection <- function(hierarchy.df,
                                                 number.of.steps = NUMBER.OF.STEPS) {
  x.minimum = min(hierarchy.df$hierarchy.log.deg) # Minimum node degree in the data frame
  x.maximum = max(hierarchy.df$hierarchy.log.deg) # Maximum node degree in the data frame
  y.minimum = min(hierarchy.df$hierarchy.log.cc)  # Minimum cluster coefficient in the data frame
  y.maximum = max(hierarchy.df$hierarchy.log.cc)  # Maximum cluster coefficient in the data frame

  # Step size
  x.step.width = (x.maximum - x.minimum)/ (number.of.steps)
  y.step.width = (y.maximum - y.minimum)/ number.of.steps

  right.boarder = x.maximum

  last.fixed.gradient <- NA
  angle.with.last.fixed <- NA

  current.points.plot <- ggplot() +
    geom_point(data = hierarchy.df, size = 0.3, aes(jitter(hierarchy.log.deg, factor = 0.5), jitter(hierarchy.log.cc, factor = 0.5))) +
    theme(legend.position="none") + xlab("log.degree") + ylab("log.clus.coef")

  color.values=colorRampPalette(c("yellow","red", "blue"))(number.of.steps+1)
  color.rectangle=colorRampPalette(c("green","blue"))(number.of.steps+1)

  color.step = 1
  for (current.step in 1:(number.of.steps)) {
    current.x.limit <- x.maximum - current.step * x.step.width
    current.right.points <- hierarchy.df[hierarchy.df$hierarchy.log.deg >=  current.x.limit, ]
    current.right.points <- current.right.points[current.right.points$hierarchy.log.deg <=  right.boarder,]
    current.left.points <- hierarchy.df[hierarchy.df$hierarchy.log.deg < current.x.limit, ]

    if (!is.null(current.right.points) && nrow(current.right.points) > 1
        &&  !is.null(current.left.points) && nrow(current.left.points) > 1
        && nrow(current.right.points)/nrow(hierarchy.df) > 0.1) {

      reg.right <- lm(hierarchy.log.cc~hierarchy.log.deg, data = current.right.points)
      reg.gradient.right <- c(1, reg.right$coefficients [2])
      var.right <- mean(resid(reg.right)^2) # variance

      reg.left <- lm(hierarchy.log.cc~hierarchy.log.deg, data = current.left.points)
      reg.gradient.left <- c(1, reg.left$coefficients [2])
      var.left <- mean(resid(reg.left)^2)

      theta <- acos(sum(reg.gradient.left*reg.gradient.right) / (sqrt(sum(reg.gradient.left * reg.gradient.left))
                                                                  * sqrt(sum(reg.gradient.right * reg.gradient.right))))
      angle <- rad2deg(theta)

      if (!is.na(angle) &&  angle >= 30.0) {
        if (!is.na(last.fixed.gradient)) {
          theta.with.last.fixed <- acos(sum(last.fixed.gradient*reg.gradient.right) / (sqrt(sum(last.fixed.gradient * last.fixed.gradient))
                                                                                       * sqrt(sum(reg.gradient.right * reg.gradient.right))))
          angle.with.last.fixed <- rad2deg(theta.with.last.fixed)
        }

        if (is.na(last.fixed.gradient) ||
            (!is.na (angle.with.last.fixed) && angle.with.last.fixed >= 30 && angle.with.last.fixed < 90)) {
          current.points.plot <- current.points.plot +
            geom_smooth(data = current.right.points, aes(hierarchy.log.deg, hierarchy.log.cc),
                        method='lm',formula=y~x, size = 0.4, se = FALSE, colour = color.values[color.step]) +
            geom_text(aes(y=0), x=current.x.limit, label=round(angle, digits = 2),
                      colour = color.values[color.step],
                      angle = 90, vjust = 1.2, size=3)

          right.boarder <- current.x.limit
          last.fixed.gradient <- reg.gradient.right
        }
      }
    }
    color.step <- color.step + 1
  }
  # letztes Stück
  current.points.plot <- current.points.plot +
    geom_smooth(data = hierarchy.df[hierarchy.df$hierarchy.log.deg < right.boarder, ],
                aes(hierarchy.log.deg, hierarchy.log.cc),
                method='lm',formula=y~x, size = 0.4, se = FALSE,
                colour = color.values[color.step])

  return(current.points.plot)
}

## Decide, which lines are important or realistic values --------------------------

#'
#' @param stepwise.data data after stepwise analysis of hierarchy at one time range of
#' one project
#'
#' @return input data set with additional flag which models potential good linear
#' regression lines
mark.realistic.values <- function(stepwise.data) {
  if (!is.null(stepwise.data) && length(stepwise.data) > 0) {
    marked.data <- stepwise.data
    marked.data$possible.regression.line <- FALSE
    marked.data$variance.smaller.threshold <- TRUE
    marked.data$angel.smaller.threshold <- TRUE
    marked.data$percent.n.higher.threshold <- TRUE
    if (!is.null(stepwise.data) && nrow(stepwise.data) > 0) {
      for (step in 1:nrow(stepwise.data)) {
        hierarchy.data.row <- marked.data[step,]
        is.realistic.value <- hierarchy.data.row$variance < MAX.VARIANCE &&
          (hierarchy.data.row$percent.n > MIN.PERC.DEVS.IN.HIERARCHICAL.PART) &&
          !is.na(hierarchy.data.row$angle) && hierarchy.data.row$angle < -MIN.ANGLE

        marked.data[step,]$variance.smaller.threshold <- hierarchy.data.row$variance < MAX.VARIANCE
        marked.data[step,]$angel.smaller.threshold <- hierarchy.data.row$angle < -MIN.ANGLE
        marked.data[step,]$percent.n.higher.threshold <- hierarchy.data.row$percent.n > MIN.PERC.DEVS.IN.HIERARCHICAL.PART
        marked.data[step,]$possible.regression.line <- is.realistic.value
      }
    }
    return(marked.data)
  }
}

#' Executed the function \code{mark.realistic.values} for all time ranges of a
#' project.
#'
#' @param hierarchy.analysis.data data after analysis of hierarchy in time ranges,
#' whole project history
#' @param only.one.value select at most one value[default: FALSE]
mark.realistic.values.of.whole.project.history <- function(hierarchy.analysis.data, only.one.value = FALSE) {
  all.marked.data <- lapply(hierarchy.analysis.data, mark.realistic.values)
  return(all.marked.data)
}

## Mark all lines as important ------------------------------------------------

#'
#' @param stepwise.data data after stepwise analysis of hierarchy at one time range of
#' one project
#'
#' @return input data set with additional flag which is always \code{TRUE}
mark.all.values <- function(stepwise.data) {
  if (!is.null(stepwise.data) && length(stepwise.data) > 0) {
    marked.data <- stepwise.data
    marked.data$possible.regression.line <- FALSE
    if (!is.null(stepwise.data) && nrow(stepwise.data) > 0) {
      marked.data$possible.regression.line <- TRUE
    }
    return(marked.data)
  }
}

#' Executed the function \code{mark.all.values} for all time ranges of a
#' project.
#'
#' @param hierarchy.analysis.data data after analysis of hierarchy in time ranges,
#' whole project history
#' @param only.one.value select at most one value[default: FALSE]
mark.all.values.of.whole.project.history <- function(hierarchy.analysis.data, only.one.value = FALSE) {
  all.marked.data <- lapply(hierarchy.analysis.data, mark.all.values)
  return(all.marked.data)
}




## Aggregate values over time per project ----------------------------------------

#' Aggregate one of the meta hierarchy parameters (created through stepwise analysis
#' of hierarchy)
#'
#'@param marked.hierarchy.data hierarchy meta data with a flag which marks the candidates
#'                             of good linear regressions
#'@param selected.data selected meta hierarchy parameter; examples are variance, angle,
#'                     percent.n, n [default: "percent.n"]
#'
#'@return list, which includes for each time step a list of the values of the meta
#'hierarchy parameter and the time steps, which are marked (can be more then one
#'value per time step)
extract.one.parameter.from.one.project <- function(marked.hierarchy.data, selected.column = "percent.n") {
  aggregated.percent.n <- lapply(marked.hierarchy.data, function(current.range.data) {
    return(current.range.data[current.range.data$possible.regression.line, selected.column])
  })
  return(aggregated.percent.n)
}

#' Plot the aggregated data of one parameter of a whole project. More than one value can
#' be used for each time step. The colors of the data points have no special meaning, they
#' only should separate the different time ranges (all points at a time step have the
#' same color).
#'
#' @param aggregated.data data which was aggregated with the function
#' \code{extract.one.parameter.from.one.project}
#' @param selected.column name of the selected column, is used to label the axis
#'                         [default: "percent.n"]
#'
#' @return plot with mean and median of the data (shown in a label)
plot.one.parameter.from.one.project <- function(aggregated.data, selected.column = "percent.n") {
  xdf <- reshape2::melt(aggregated.data)

  # average value of all realistic values
  average.value <- mean(xdf$value)
  median.value <- median(xdf$value)
  string.mean.and.median <- paste(c("mean: ", round(average.value, digits = 2),
                                    ", median: ", round(median.value, digits = 2)),
                                  collapse = " ")

  # the different colors are only used to seperat the different time steps
  result.plot <- ggplot(data = xdf, aes(L1, value, color=(L1 %% 5))) + geom_point() +
    scale_colour_gradientn(colours=rainbow(5)) +
    xlab("time steps") + ylab(selected.column)

  # add mean and median to the plot
  if(selected.column == "percent.n") {
    result.plot <- result.plot + ylim(0,1) +
      geom_label(aes(y=1), x=15, colour="black", label=string.mean.and.median, size=3) +
      geom_hline(yintercept = MIN.PERC.DEVS.IN.HIERARCHICAL.PART, colour="black", linetype="dotted") + geom_smooth(colour="blue", level=0)
  } else if (selected.column == "angle") {
    result.plot <- result.plot +
      geom_label(aes(y=-25), x=15, colour="black", label=string.mean.and.median, size=3) +
      geom_hline(yintercept = -MIN.ANGLE, colour="black", linetype="dotted") + geom_smooth(colour="blue", level=0)
  } else if (selected.column == "variance") {
    result.plot <- result.plot +
      geom_label(aes(y=0.2), x=15, colour="black", label=string.mean.and.median, size=3) +
      geom_hline(yintercept = MAX.VARIANCE, colour="black", linetype="dotted") + geom_smooth(colour="blue", level=0)
  } else {
    result.plot <- result.plot +
      geom_label(aes(y=1), x=15, colour="black", label=string.mean.and.median, size=3)
  }

  return(result.plot)
}

#' Compute mean, median, max, min of the data regarding one parameter of the hierarchy
#' analysis.
#'
#' @param aggregated.data data of one parameter
#'
#' @return tupel with min, max, mean, median of the values
aggregation.analysis.one.parameter.from.one.project <- function(aggregated.data) {
  xdf <- reshape2::melt(aggregated.data)

  # average value of all realistic values
  average.value <- mean(xdf$value)
  median.value <- median(xdf$value)
  min.value <- min(xdf$value)
  max.value <- max(xdf$value)

  return(data.frame("min" = min.value, "max" = max.value, "mean" = average.value,
                    "median" = median.value))
}

## Compute standard error and mean to normalize --------------------------

#' Compute the standard error and the mean of each parameter over a whole project and
#' all iteration steps per time range.
#'
normalize.criteria <- function(hierarchy.data) {

  #hierarchy.data <- hierarchy.data[!unlist(lapply(hierarchy.data, is.null))] # remove NULL values
  hierarchy.data <- plyr::compact(hierarchy.data)

  # percent.n
  percent.n.standard.error.and.mean <- compute.standard.errors(hierarchy.data,
                                                               selected.column = "percent.n")

  hierarchy.data.with.normalized.percent.n <-
    lapply(hierarchy.data,
           function(current.range) {
             current.range$normalized.percent.n <- (current.range$percent.n - percent.n.standard.error.and.mean$mean.value )/   percent.n.standard.error.and.mean$sigma
             return(current.range)
    })

  # variance, standard error?
  variance.standard.error.and.mean <- compute.standard.errors(hierarchy.data.with.normalized.percent.n,
                                                              selected.column = "variance")

  hierarchy.data.with.normalized.variance <-
    lapply(hierarchy.data.with.normalized.percent.n,
           function(current.range) {
             # minimize
             current.range$normalized.variance <- -(current.range$variance -  variance.standard.error.and.mean$mean.value )/ variance.standard.error.and.mean$sigma
             return(current.range)
    })

  # angle
  angle.standard.error.and.mean <- compute.standard.errors(hierarchy.data.with.normalized.variance,
                                                           selected.column = "angle")

  hierarchy.data.with.normalized.angle <-
    lapply(hierarchy.data.with.normalized.variance,
           function(current.range) {
             # minimize
             current.range$normalized.angle <- -(current.range$angle -  angle.standard.error.and.mean$mean.value )/ angle.standard.error.and.mean$sigma
             return(current.range)
    })
  hierarchy.data.with.weighted.sum <-
    lapply(hierarchy.data.with.normalized.angle,
           function(current.range) {
             current.range$weighted.sum = current.range$normalized.angle+
               current.range$normalized.variance+
               3*current.range$normalized.percent.n
             return(current.range)
    })

  return(hierarchy.data.with.weighted.sum)
}

#' Compute standard error and mean
compute.standard.errors <- function(hierarchy.data, selected.column = "percent.n") {
  selected.data <- lapply(hierarchy.data, function(current.range.data) {
    return( current.range.data[, selected.column])
  })

  selected.data <- unlist(selected.data)
  selected.data<-selected.data[!is.na(selected.data)]
  mean.value <- mean(selected.data)
  sigma <- sd(selected.data)
  return(data.frame("mean.value"=mean.value, "sigma" = sigma))
}

#' Maximize the sum of the criteria
#'
maximize.criteria.sum <- function(hierarchy.with.normalized.criteria, previous.selection = TRUE) {
  if(!previous.selection) {
    selected.developer.groups <- lapply(hierarchy.with.normalized.criteria, function(hierarchical.range){
      if (is.null(hierarchical.range) || anyNA(hierarchical.range$angle)) {
        logging::loginfo("Data frame is null.")
        return(NULL)
      }
      hierarchical.range$selected.value <- FALSE

      hierarchical.range[max(hierarchical.range$normalized.angle +
                               hierarchical.range$normalized.variance +
                               3*hierarchical.range$normalized.percent.n) # TODO: move code duplication of weighted sum
                         == hierarchical.range$normalized.angle+
                           hierarchical.range$normalized.variance+
                           3*hierarchical.range$normalized.percent.n, "selected.value"] = TRUE
      return(hierarchical.range)
    })
  } else {
    candidate.developer.groups <- lapply(hierarchy.with.normalized.criteria, function(hierarchical.range){
      return(hierarchical.range[hierarchical.range$possible.regression.line,])
    })

    selected.developer.groups <- lapply(candidate.developer.groups, function(hierarchical.range){

      if (nrow(hierarchical.range) == 0) {
        return(NULL)
      }
      hierarchical.range$selected.value <- FALSE
      hierarchical.range[max(hierarchical.range$normalized.angle+
                               hierarchical.range$normalized.variance+
                               3*hierarchical.range$normalized.percent.n) # TODO: move code duplication of weighted sum
                         == hierarchical.range$normalized.angle+
                           hierarchical.range$normalized.variance+
                           3*hierarchical.range$normalized.percent.n, "selected.value"] = TRUE
      return(hierarchical.range)
    })

  }
  return(selected.developer.groups)
}

plot.hierarchy.only.for.selected.group <- function(hierarchy.df,
                                                   hierarchy.data.with.normalized.criteria,
                                                   number.developers,
                                                   number.of.steps = NUMBER.OF.STEPS) {

  logging::loginfo("Start, plotting of hierarchy with one selected line")
  hiera.df <- hierarchy.df[[1]]
  x.minimum = min(hiera.df$hierarchy.log.deg) # Minimum node degree in the data frame
  x.maximum = max(hiera.df$hierarchy.log.deg) # Maximum node degree in the data frame
  y.minimum = min(hiera.df$hierarchy.log.cc)  # Minimum cluster coefficient in the data frame
  y.maximum = max(hiera.df$hierarchy.log.cc)  # Maximum cluster coefficient in the data frame

  # Step size
  x.step.width = (x.maximum - x.minimum)/ (number.of.steps)
  y.step.width = (y.maximum - y.minimum)/ number.of.steps

  current.points.plot <- ggplot() +
    geom_point(data = hiera.df, size = 0.5, aes(jitter(hierarchy.log.deg, factor = 0.7), jitter(hierarchy.log.cc, factor = 0.7))) +
    theme_light(base_size = 25) + theme(legend.position="none") +
    xlab(paste("log degree \n 0 contacts: ", round(number.developers$num.deg.zero.percent, digits = 1),
               " %, 1 contact: ", round(number.developers$num.deg.one.percent, digits = 1),
               " %, >1 contact: ", round(number.developers$num.deg.greater.one.percent, digits = 1),
               " %", sep="")) +
           ylab("log cluster coefficient") +
    scale_x_continuous( labels=scaleFUN) + scale_y_continuous(labels=scaleFUN)


  if (!is.null(hierarchy.data.with.normalized.criteria[[1]])){
    hierarchy.data.w.n.c <- hierarchy.data.with.normalized.criteria[[1]]
    selected.line <- hierarchy.data.w.n.c[hierarchy.data.w.n.c$selected.value, ]
    current.x.limit <- x.maximum - selected.line$cur.step[1] * x.step.width
    current.points <- hiera.df[hiera.df$hierarchy.log.deg >=  current.x.limit, ]

    if (nrow(current.points) > 2) {
      current.points.plot <- current.points.plot +
        geom_smooth(data = current.points, aes(hierarchy.log.deg, hierarchy.log.cc),
                    method='lm',formula=y~x, size = 0.4, se = FALSE, colour = "red") +
        geom_vline(xintercept = current.x.limit, size = 0.9, color = "azure4", linetype = "dashed") +
        geom_text(aes(y=0), x=current.x.limit, label=selected.line$cur.step[1],
                  colour = "azure4",
                  angle = 90, vjust = 1.2, size=3)
    } else {
      print("Too few data points.")
    }
  }

  return(current.points.plot)
}


## Aggregation of one  value

select.candidate.partial.hierarchy <- function(analysed.hierarchy.with.selection) {
  selected.candidates <- lapply(analysed.hierarchy.with.selection, function(all.candicates.per.range) {
    if (!is.null(all.candicates.per.range)) {
      cut.hierarchy.data <- subset(all.candicates.per.range, select = -c(gradient, perc.n.diff, vari.diff, angle.diff, cur.step))
      ## TODO: Problem: aktueller Schritt ist weg
      all.candicates.per.range$duplicated <- duplicated(cut.hierarchy.data)
      return(all.candicates.per.range[all.candicates.per.range$selected.value & !all.candicates.per.range$duplicated, ])
    }
  })
  return(selected.candidates)
}

plot.parameters.from.one.project <- function(data, developer.numbers, output.file) {

  # node degree
  subdata <- developer.numbers[,c('num.deg.greater.one', 'num.deg.one', 'num.deg.zero', 'range.number')]
  df <- melt(subdata, id.vars = 'range.number')

  plot.node.degrees <- ggplot(data=df, aes(x=range.number, y=value, group=variable, fill=variable))  + geom_area(position="fill") +
    figureargs + theme(legend.position="bottom", axis.title.y=element_blank()) +
    theme(legend.title=element_blank()) + ylab(" ") + xlab("time steps") +
    scale_y_continuous(labels= function(x) paste0(100*x,'%')) +
    scale_fill_manual(values = c("aliceblue", "lightblue",   "darkblue"),
                      name  ="", breaks=c("num.deg.zero", "num.deg.one", "num.deg.greater.one"),
                      labels=c("Perc. 0 contacts  ", "Perc. 1 contact  ", "Perc. >1 contact  "))

  # use developer.number
  average.value <- mean(developer.numbers[["num.dev"]])
  median.value <- median(developer.numbers[["num.dev"]])
  string.mean.and.median <- paste(c("mean:", round(average.value, digits = 2),
                                    ", median:", round(median.value, digits = 2),
                                    ", first:", developer.numbers$num.dev[ 1] ,
                                    ", last:", developer.numbers$num.dev[nrow(developer.numbers)],
                                    ", \n range:(", min(developer.numbers$num.dev),
                                    ",", max(developer.numbers$num.dev),
                                    "), entire number:", developer.numbers$entire.num.dev[1]
                                    ),
                                  collapse = "")

  developer.plot <- ggplot(data = developer.numbers, aes(x=range.number, y=num.dev, color="black")) +
    geom_point(colour="black") + geom_smooth(colour="darkblue", level=0) +
    xlab(paste("time steps", "\n" ,string.mean.and.median, sep=" ")) + ylab("number of developers") +  theme(legend.position="none") + figureargs #+
    #geom_label(aes(y=400), x=15, colour="black", label=string.mean.and.median, size=8)


  plot.one.parameter <- function(xdf, selected.column = "percent.n") {
    # average value of all realistic values
    average.value <- mean(xdf[[selected.column]])
    median.value <- median(xdf[[selected.column]])
    string.mean.and.median <- paste(c("mean:", round(average.value, digits = 2),
                                      ", median:", round(median.value, digits = 2)),
                                    collapse = "")

    # the different colors are only used to separate the different time steps
    result.plot <- ggplot(data = xdf, aes(range.number, xdf[[selected.column]], color="black")) + geom_point(colour="black") +
      #scale_colour_gradientn(colours=rainbow(5)) +
      xlab(paste("time steps", "\n" ,string.mean.and.median, sep=" ")) + ylab(selected.column)  + figureargs +  theme(legend.position="none")

    # add mean and median to the plot
    if(selected.column == "percent.n") {
      result.plot <- result.plot + ylim(0,1) +
        #geom_label(aes(y=1), x=15, colour="black", label=string.mean.and.median, size=8) +
        geom_hline(yintercept = MIN.PERC.DEVS.IN.HIERARCHICAL.PART, colour="black", linetype="dotted") + geom_smooth(colour="darkblue", level=0) +
        scale_y_continuous(labels= function(x) paste0(100*x,'%'))
    } else if (selected.column == "angle") {
      result.plot <- result.plot +
        #geom_label(aes(y=-25), x=15, colour="black", label=string.mean.and.median, size=8) +
        geom_hline(yintercept = -MIN.ANGLE, colour="black", linetype="dotted") + geom_smooth(colour="darkblue", level=0)
    } else if (selected.column == "variance") {
      result.plot <- result.plot +
        #geom_label(aes(y=0.2), x=15, colour="black", label=string.mean.and.median, size=8) +
        geom_hline(yintercept = MAX.VARIANCE, colour="black", linetype="dotted") + geom_smooth(colour="darkblue", level=0) +
        ylab("variance")
    } else if (selected.column == "number.all.developer") {
      result.plot <- result.plot +
        #geom_label(aes(y=-5), x=15, colour="black", label=string.mean.and.median, size=8) +
        geom_hline(yintercept = 2, colour="black", linetype="dotted") + geom_smooth(colour="darkblue", level=0)
    } else {
      result.plot <- result.plot +
        geom_label(aes(y=1), x=15, colour="black", label=string.mean.and.median, size=8)
    }

    return(result.plot)
  }


  plots <- list()
  plots[[1]] <- developer.plot
  plots[[2]] <- plot.node.degrees
  plots[[3]] <- plot.one.parameter(data, "percent.n")
  plots[[4]] <- plot.one.parameter(data, "angle")
  plots[[5]] <- plot.one.parameter(data, "variance")


  pdf(file=output.file, height= 5*6, width = 10)
  do.call("grid.arrange", c(plots, ncol=1, newpage=TRUE))
  dev.off()
}

## Validation of the algorithm -------------------------------

extract.number.of.selected.candidate.group <- function(data) {
  selected.numbers <- list()
  for (range.data.number in 1:length(data)) {
    range.data <- data[[range.data.number]]
    if (!is.null(range.data)){
      range.name <- names(data[range.data.number][1])
      selected.numbers[[range.data.number]] <- data.frame("range.num"= range.data.number, "range" = range.name,
                     "cur.step" = range.data[range.data$selected.value, "cur.step"])
    } else {
      selected.numbers[[range.data.number]] <- null
    }
  }
  return(selected.numbers)
}


## Function to mark hierarchical structure developers ---------------------------

#' The function compute the division into a hierarchical part and a
#' non-hierarchical part
#'
#' @param hierarchy list of data frames including for each developer
#'                  node degree and cluster coefficient
#' @param number.of.steps number of candidate buckets for hierarchical
#'                  group
#' @param hierarchy.data.with.normalized.criteria hierarchy with further parameters
#' @param net.name name of the range, which should be analyzed
#' @return hierarchy with an additional \code{part.of.hierarchy} with
#' values \code{TRUE} for developers in the hierarchy and \code{FALSE} else.
#'
compute.division.into.hierarchical.and.non.hierachical.part <- function(hierarchy,
                                                                        number.of.steps = NUMBER.OF.STEPS,
                                                                        hierarchy.data.with.normalized.criteria,
                                                                        net.name) {
  if (is.null(hierarchy)) {
    return()
  }

  x.minimum = min(hierarchy$hierarchy.log.deg) # Minimum node degree in the data frame
  x.maximum = max(hierarchy$hierarchy.log.deg) # Maximum node degree in the data frame
  y.minimum = min(hierarchy$hierarchy.log.cc)  # Minimum cluster coefficient in the data frame
  y.maximum = max(hierarchy$hierarchy.log.cc)  # Maximum cluster coefficient in the data frame

  logging::logdebug(paste("compute.division.into.hierarchical.and.non.hierachical.part: net.name: ",
                          net.name, ",   x.maximum: ", x.maximum))

  # Step size
  x.step.width = (x.maximum - x.minimum)/ number.of.steps
  y.step.width = (y.maximum - y.minimum)/ number.of.steps

  # identifying the nodes in the hierarchically structured part
  hierarchy.data.w.n.c <- hierarchy.data.with.normalized.criteria[[net.name]]
  selected.line <- hierarchy.data.w.n.c[hierarchy.data.w.n.c$selected.value, ]
  current.x.limit <- x.maximum - selected.line$cur.step[1] * x.step.width
  current.points <- hierarchy[hierarchy$hierarchy.log.deg >=  current.x.limit, ]

  # mark node of the hierarchy
  hierarchy$part.of.hierarchy <- NA
  hierarchy[hierarchy$hierarchy.log.deg >=  current.x.limit, "part.of.hierarchy"] <- TRUE
  hierarchy[hierarchy$hierarchy.log.deg <  current.x.limit, "part.of.hierarchy"] <- FALSE

  return(hierarchy)
}


## Function to compute hierarchy and mark developers ---------------

#' The function computes the hierarchy of a network and marks the developers
#' in the hierarchy.
#'
#' @param networks list of networks
#' @param number.of.buckets number of buckets for the candidate groups, [default: 20]
#' @return list of data frames including the developers' names and a boolean values
#' \code{part.of.hierarchy}, weather the developer is part of the hierarchy or not.
#'
compute.hierarchical.structured.developers <- function(networks, number.of.buckets = NUMBER.OF.STEPS) {

  # remove networks with less the two nodes
  networks <- networks[lapply(networks, vcount) >1]

  # compute node degree and cluster coefficient for all developers in a network
  node.degrees.and.cluster.coefficients <- lapply(names(networks), function(range) {
    hierarchy <- metrics.hierarchy(networks[[range]])
    finite.hier <- subset(hierarchy, !(is.nan(cc) | is.infinite(log.cc))) ## revome NaN and infinite
    if (nrow(finite.hier) != 0) {
      hier.df <- data.frame("range" = range, "hierarchy" = finite.hier)
      return(hier.df)
    } else {
      return(NULL)
    }
  })

  names(node.degrees.and.cluster.coefficients) <- names(networks)
  node.degrees.and.cluster.coefficients <- plyr::compact(node.degrees.and.cluster.coefficients)

  candidate.groups.metadata <- lapply(node.degrees.and.cluster.coefficients,
                                      function(one.rang.nd.and.cc) {
                                        return(compute.hierarchical.parts(
                                          one.rang.nd.and.cc,
                                          number.of.steps = number.of.buckets))
                               })

  candidate.groups.metadata <- plyr::compact(candidate.groups.metadata) # remove NULL values

  candidate.groups.metadata <- mark.realistic.values.of.whole.project.history(
    candidate.groups.metadata)

  candidate.groups.with.normalized.metadata <-
    normalize.criteria(candidate.groups.metadata)

  optimal.candidate.metadata <- maximize.criteria.sum(
    candidate.groups.with.normalized.metadata,
    previous.selection = TRUE)

  marked.hierarchical.structured.developers <- lapply(names(networks),function(network.name){
    return(compute.division.into.hierarchical.and.non.hierachical.part(
      node.degrees.and.cluster.coefficients[[network.name]],
      number.of.steps = number.of.buckets,
      optimal.candidate.metadata,
      net.name = network.name
    ))
  })

  return(marked.hierarchical.structured.developers)
}
